/*
 * LICTiA1FController.cpp
 *
 *  Created on: 2020/05/07
 *      Author: Tsuyoshi Anazawa
 *
 *     Outline: LICTiA1Fのドア, センサ制御プログラム
 *
 */
#include <cnoid/SimpleController>
#include <cnoid/Joystick>
#include <cnoid/RangeSensor>
#include <cnoid/ConnectionSet>
#include <fmt/format.h>
#include <iostream>
#include <sstream>

using namespace std;
using namespace cnoid;
using fmt::format;

class LICTiA1FController : public SimpleController
{
	// Door格納変数
	vector<Link*> door;
	// push/pull型のDoorの数
	static const int DOORNUM = 27;
	static const int SLIDEDOORNUM = 9;
	// Doorの名称
	string doorNames[DOORNUM] = {/* 個室	,        会議室3,      会議室2 */
									"Pr_door_0", "Mr3_door_0", "Mr2_door_0",
								 /*	会議室1,      ホワイトボード, センター長室 */
									"Mr1_door_0", "Wbr_door_0", "Mr_door_0",
								 /* 事務室,               前室,         検証室 */
									"Clerks_room_door_0", "Fc_door_0", "Vr_door_0",
								 /*	サイバーセキュリティウォールーム, 倉庫3右,               倉庫3左 */
									"Cyber_security_room_door_0",  "Storehouse_3_door_0", "Storehouse_3_door_1",
								 /* 倉庫2                倉庫1,              女子WC, */
									"Storehouse2_door", "Storehouse1_door", "Woman_wc_door",
								 /*	男子WC         階段,         裏口 */
									"Man_wc_door", "Stair_door", "Be_door_0",
								 /* 添字が18以降は Slide door  */
								 /* 多目的WC,                エレベータ内側左,     エレベータ内側右 */
									"Accessible_wc_door", "ELV_DOOR_INSIDE_L", "ELV_DOOR_INSIDE_R",
								 /* エレベータ外側左,      エレベータ外側右,      風除室外側左 */
									"ELV_DOOR_OUTSIDE_L", "ELV_DOOR_OUTSIDE_R", "Wb_fl_door",
								 /* 風除室外側右,  風除室内側左,  風除室内側右 */
									"Wb_fr_door", "Wb_ll_door", "Wb_lr_door" };
	double slideDistance[SLIDEDOORNUM] = {/* 多目的WC,	エレベータ内側左,		エレベータ内側右  */
											 0.795, 	0.445,				0.4491,
										  /* エレベータ外側左,	エレベータ外側右,		風除室外側左 */
											 0.445,				0.4491,				0.6670,
										  /* 風除室外側右,	風除室内側左,	風除室内側右 */
											 0.667,			0.687,			0.687 };
	// Body情報格納変数
	Body* body;
	// アクチュエーションモード
	Link::ActuationMode actuationMode;
	double qprev[DOORNUM];
	double qref[DOORNUM];
	double dt;
	// Range sensor情報
	RangeSensorPtr rangeSensor[6];
	std::shared_ptr<const RangeSensor::RangeData> prevRangeData[6];
	char id[13];
	string deviceName;
	double dis[3];	// 0:風除室の外側自動ドア, 1:風除室の内側自動ドア, 2:エレベーター
	bool openFlg;
	// 現在のスコープ限定の接続
	ScopedConnectionSet rangeConnections;

public:
	virtual bool initialize(SimpleControllerIO* io) override
	{
		ostream& os = io->os();
		// Bodyの取得
		body = io->body();

		// アクチュエーションモードの初期化
		actuationMode = Link::JOINT_TORQUE;
		// コントローラオプションの取得
		string option = io->optionString();

		if(!option.empty()){
			// コントローラオプションが空の場合
			if(option == "velocity" || option == "position"){
				// velocity または positionの場合
				actuationMode = Link::JOINT_VELOCITY;
			} else if(option == "torque"){
				// torqueの場合
				actuationMode = Link::JOINT_TORQUE;
			} else{
				// 上記以外の場合
				os << format("Warning unknown option \"{}\".", option) << endl;
			}
		}

		// push/pullのDoor
		for(int i = 0; i < DOORNUM; ++i){
			// 配列にドアリンクを格納
			door.push_back(io->body()->link(doorNames[i]));

			if(!door[i]){
				// リンクが取得できない場合
				os << format("{0} of {1} is not found.", doorNames[i], body->name()) << endl;
				return false;
			}

			// リンクのアクチュエーションモードの設定
			door[i]->setActuationMode(actuationMode);
			// ドアへの出力を有効化
			io->enableIO(door[i]);
			qref[i] = qprev[i] = door[i]->q();
		}

		// 接続の切断
		rangeConnections.disconnect();
		for(int i = 0; i < body->devices<RangeSensor>().size(); ++i){
			// レンジセンサ番号を設定
			sprintf(id, "%d", i);
			// レンジセンサ名を格納
			deviceName = "RangeSensor_" + (string)id;
			// レンジセンサを取得
			rangeSensor[i] = body->findDevice<RangeSensor>(deviceName);

			if(!rangeSensor[i]){
				// 該当するレンジセンサが存在しない場合
				os << format("{} is not found.", deviceName) << endl;
				return false;
			}

			// レンジセンサからの入力を有効化
			io->enableInput(rangeSensor[i]);
			initializeRangeSensor();
			// センサの状態が変わった場合、onRangeSensorStateChanged()を呼び出す
			rangeConnections.add(rangeSensor[i]->sigStateChanged().connect(
					[&](){ onRangeSensorStateChanged(); }));
		}

		dt = io->timeStep();
		openFlg = false;

		return true;
	}

	virtual bool control() override
	{
		for(int i = 0; i < DOORNUM; ++i){
			double u;

			if(dt == 0.0){
				return false;
			}

			double dq_current = (door[i]->q() - qprev[i]) / dt;
			double dq_ref = 0.0;

			if(i < (DOORNUM - SLIDEDOORNUM)){
				// PD制御定数
				const double Kp = 25.0;
				const double Kd = 25.0;

				if((i <= 3 and door[i]->q() > round(M_PI * 100) / 200 + 0.01)
						or (i >= 4 and door[i]->q() < 0.0)){
					qref[i] += dt * 0.05;
				} else if((i <= 3 and door[i]->q() < 0.1)
						or (i >= 4 and door[i]->q() > 0.1 and door[i]->q() < round(M_PI * 100) / 200 - 0.1)){
					qref[i] += dt * (-0.05);
				} else {
					if(i <= 3){
						qref[i] = round(M_PI * 100) / 200;
					} else {
						qref[i] = 0.0;
					}
				}

				u = Kp * (qref[i] - door[i]->q()) + Kd * (dq_ref - dq_current);

			} else if(i >= (DOORNUM - SLIDEDOORNUM)){
				// PD制御定数
				double Kp = 100.0;
				double Kd = 350.0;
				// slide式のドアの場合
				if(i == 18){
					if(door[i]->q() < 0.0 or door[i]->q() >= slideDistance[0]){
						qref[i] += 0.05 * dt;
					} else if(door[i]->q() > 0.1 and door[i]->q() < slideDistance[0] - 0.1){
						qref[i] += (-0.05) * dt;
					} else {
						qref[i] = 0.0;
					}
					Kp = 200.0;
					Kd = 250.0;

					u = Kp * (qref[i] - door[i]->q()) + Kd * (dq_ref - dq_current);
				} else if((i >= 19 and i <= 22 and dis[2] < 1.7)
						or (i >= 23 and i <= 24 and dis[0] < 1.7)
						or (i >= 25 and i <= 26 and dis[1] < 1.7)){
					// 自動ドア かつ レンジセンサで取得した距離 < 1.7 の場合
					if(!openFlg){
						// 開閉フラグがfalseの場合
						u = Kp * qref[i] + Kd * (dq_ref - dq_current);
					} else {
						// 上記以外の場合
						if(door[i]->q() >= slideDistance[i - (DOORNUM - SLIDEDOORNUM)] - 0.1){
							qref[i] = slideDistance[i - (DOORNUM - SLIDEDOORNUM)] - 0.1;
						} else {
							qref[i] += 0.5 * dt;
						}
						u = Kp * (qref[i] - door[i]->q()) + Kd * (dq_ref - dq_current);
					}
				} else if((i >= 19 and i <= 22 and dis[2] >= 1.7)
						or (i >= 23 and i <= 24 and dis[0] >= 1.7)
						or (i >= 25 and i <= 26 and dis[1] >= 1.7)){
					// 自動ドア かつ レンジセンサで取得した距離 >= 1.7 の場合
					if(openFlg){
						// 開閉フラグがfalseの場合
						u = Kp * qref[i] + Kd * (dq_ref - dq_current);
					} else {
						// 上記以外の場合
						if(door[i]->q() <= 0.1){
							qref[i] = 0.0;
						} else {
							if(qref[i] > 0.0){
								qref[i] = 0.0;
							} else {
								qref[i] += (-0.5) * dt;
							}
						}
						u = Kp * (qref[i] - door[i]->q()) + Kd * (dq_ref - dq_current);
					}
				}
			} else{
				// 上記以外の場合
				u = 0.0;
			}
			door[i]->u() = u;
			qprev[i] = door[i]->q();
		}
		return true;
	}

	virtual void stopSimulation()
	{
//		for(auto sensor : rangeSensor){
//			sensor.reset();
//		}
	}

	void initializeRangeSensor()
	{
		for(int i = 0; i < 3; ++i){
			dis[i] = 100.0;

		}
	}

	void onRangeSensorStateChanged()
	{
		// レンジセンサの数分処理を繰り返す
		for(int i = 0; i < body->devices<RangeSensor>().size(); ++i){
			if(rangeSensor[i]->sharedRangeData() != prevRangeData[i]){
				// レンジセンサの値が前回値と異なる場合
				// 前回値を現在値で更新
				prevRangeData[i] = rangeSensor[i]->sharedRangeData();
				// レンジセンサの値を取得
				const RangeSensor::RangeData& src = rangeSensor[i]->constRangeData();

				// 取得したレンジセンサの値分処理を繰り返す
				for(size_t j = 0; j < src.size(); ++j){
					if(i >= 0 and i <= 1){
						// 外側の自動ドアの場合
						if(src[j] < 1.7){
							// レンジセンサの値 < 1.7 の場合
							dis[0] = src[j];
							openFlg = true;
							break;
						} else {
							// 上記以外の場合
							dis[0] = src[j];
							openFlg = false;
						}
					} else if(i >= 2 and i <= 3){
						// 内側の自動ドアの場合
						if(src[j] < 1.7){
							// レンジセンサの値 < 1.7 の場合
							dis[1] = src[j];
							openFlg = true;
							break;
						} else {
							// 上記以外の場合
							dis[1] = src[j];
							openFlg = false;
						}
					} else if(i >= 4 and i <= 5){
						// エレベータードアの場合
						if(src[j] < 1.7){
							// レンジセンサの値 < 1.7 の場合
							dis[2] = src[j];
							openFlg = true;
							break;
						} else {
							// 上記以外の場合
							dis[2] = src[j];
							openFlg = false;
						}
					}
				}
			}
			if(openFlg){
				// 開閉フラグ = true の場合
				break;
			}
		}
	}
};

CNOID_IMPLEMENT_SIMPLE_CONTROLLER_FACTORY(LICTiA1FController)
