import itertools
from enum import Enum
import math
import yaml
import logging
import quaternion
import numpy as np

logger = logging.getLogger("VIEWER").getChild(__name__)

SEMI_STATIC_OBJECT_MODEL_TBL = ["Table", "Chair", "Partition"]
DYNAMIC_OBJECT_MODEL_TBL = ["TurtleBot3", "Megarover", "Person", "Trolley", "End"]
# DYNAMIC_OBJECT_MODEL_TBL = ["TurtleBot3", "Megarover", "Trolley", "End"]

DELETE_MODEL_NAME_TBL = list(itertools.chain.from_iterable([DYNAMIC_OBJECT_MODEL_TBL, SEMI_STATIC_OBJECT_MODEL_TBL]))
PARENT_FREE_SPACE = "FreeSpace"
MODEL_HEIGHT_TBL = [
    {"name": "TurtleBot3", "Height": 0.05},
    {"name": "Megarover", "Height": 0.05},
    {"name": "Table", "Height": 0.72},
    {"name": "Chair", "Height": 0.4},
    {"name": "Partition", "Height": 0.3},
    {"name": "Person", "Height": 0.7},
    {"name": "Trolley", "Height": 0.2},
    {"name": "End", "Height": -1.0}
]


class UriModel(Enum):
    # TurtleBot3 = "${SHARE}/cnoid_turtlebot_bringup/model/Waffle_pi_LiDAR_2d.body"
    Megarover = "${SHARE}/cnoid_turtlebot_bringup/model/Waffle_pi_LiDAR_2d.body"
    Person = "${SHARE}/LICTiA/model/Person.body"
    Table = "${SHARE}/LICTiA/model/Table-120x120.body"
    Chair = "${SHARE}/LICTiA/model/Office_chair.body"
    Partition = "${SHARE}/LICTiA/model/Partition.body"
    Trolley = "${SHARE}/LICTiA/model/Trolley.body"
    End = "${SHARE}/LICTiA/model/dummy.body"

    @classmethod
    def get_body(cls, name):
        body = None
        for val in cls:
            if val.name.lower() == name.lower():
                body = val.value
                break
        return body


class CreateYaml:

    def __init__(self):
        pass

    @staticmethod
    def read(file_path):
        """
        Read YAML file
        :param file_path: file path
        :return: Read YAML data
        """
        yaml_data = None
        try:
            # Read yaml file.
            with open(file_path, 'r') as f:
                yaml_data = yaml.safe_load(f)
        except Exception as e:
            logger.error("Unable to read YAML file")
        return yaml_data

    @staticmethod
    def del_element(yaml_data):
        """
        Delete environment elements in YAML data
        :param yaml_data: YAML data
        :return: YAML data with environment elements removed
        """
        for index in reversed(range(len(yaml_data["Environment"]))):
            element_name = yaml_data["Environment"][index]["name"]
            parent = yaml_data["Environment"][index]["parent"]
            del_model_name = [model_name.lower() for model_name in DELETE_MODEL_NAME_TBL if
                              model_name.lower() in element_name.lower()]
            if 0 < len(del_model_name) and parent == PARENT_FREE_SPACE:
                del yaml_data["Environment"][index]
        return yaml_data

    @classmethod
    def set_element(cls, yaml_data, object_a):
        """
        Place new environment element in YAML data
        :param yaml_data: YAML data
        :param object_a: Object data to be placed
        :return: YAML data with new environment elements placed
        """
        table_coord_a = [[object_a[index]["coord"]["x"], object_a[index]["coord"]["y"]] for index in range(len(object_a)) if object_a[index]["name"].lower() in "table"]

        for index in range(len(object_a)):

            set_name = "{}-{}".format(object_a[index]["name"], index+1)
            set_parent = PARENT_FREE_SPACE
            set_uri = UriModel.get_body(object_a[index]["name"])
            if set_uri is None:
                continue
            set_x = object_a[index]["coord"]["x"]
            set_y = object_a[index]["coord"]["y"]
            if object_a[index]["name"].lower() not in "table":
                for table_coord in table_coord_a:
                    set_x, set_y = cls.adjust_coordinates(table_coord[0], table_coord[1], set_x, set_y, 0.6)
            # set_translation = [object_a[index]["coord"]["x"], object_a[index]["coord"]["y"], cls.get_object_coord_z(set_name)]
            set_translation = [set_x, set_y, cls.get_object_coord_z(set_name)]
            quat = np.quaternion(-object_a[index]["quaternion"]["w"], object_a[index]["quaternion"]["x"], object_a[index]["quaternion"]["y"], object_a[index]["quaternion"]["z"])
            euler = quaternion.as_euler_angles(quat)
            deg = float(np.rad2deg(float(euler[2])))*2
            if deg > 0:
                deg = deg % 360
            else:
                deg = (deg % 360) * -1
            deg = 0.0
            set_rotation = [0.0, 0.0, deg]
            print("{}: {}".format(object_a[index]["name"],  set_rotation))
            set_data = {"name": set_name, "parent": set_parent, "uri": set_uri, "translation": list(set_translation), "rotation": set_rotation}
            yaml_data["Environment"].append(set_data)
        else:
            set_data = {"name": "end", "parent": PARENT_FREE_SPACE, "uri": UriModel.get_body("End"), "translation": [0,0,-1], "rotation": [0,0,0]}
            yaml_data["Environment"].append(set_data)
        return yaml_data

    @staticmethod
    def get_object_coord_z(name):
        """
        Change the height of each placed object
        :param name: Object name
        :return: height
        """
        set_z = 0
        height = [model["Height"] for model in MODEL_HEIGHT_TBL if model["name"].lower() in name.lower()]
        if 0 < len(height):
            set_z = height[0]
        return set_z

    @staticmethod
    def write(yaml_data, file_path):
        """
        Write YAML data
        :param yaml_data: YAML data
        :param file_path: Write destination
        :return: Write result
        """
        result = False
        try:
            # Read yaml file.
            with open(file_path, 'w') as f:
                yaml.dump(yaml_data, f, default_flow_style=False)
                result = True
        except Exception as e:
            logger.error("Unable to output YAML file")
        return result

    @staticmethod
    def update(object_a, read_path, write_path):
        """
        Create a YAML file with updated positions of quasi-static objects
        :param object_a: Object data to be placed
        :param read_path: Read YAML file path
        :param write_path: Write YAML file path
        :return: Update result
        """
        result = False
        try:
            yaml_data = CreateYaml.read(read_path)
            if yaml_data is None:
                return result
            yaml_data = CreateYaml.del_element(yaml_data)
            yaml_data = CreateYaml.set_element(yaml_data, object_a)
            if CreateYaml.write(yaml_data, write_path):
                result = True
        except Exception as e:
            logger.error(e)
        return result

    @staticmethod
    def adjust_coordinates(x, y, x1, y1, h):
        distance = math.sqrt((x1 - x) ** 2 + (y1 - y) ** 2)

        if distance <= h:
            # x1, y1͈͓̏ꍇAahOɍWύX
            step = 0.1
            x_diff = x1 - x
            y_diff = y1 - y

            # ł邾Ƃ̍Wɋ߂lŕԂ
            new_x = x1 + step * round(x_diff / step)
            new_y = y1 + step * round(y_diff / step)

            return new_x, new_y
        else:
            # ͈͊Ȍꍇ͂̂܂܍WԂ
            return x1, y1

if __name__ == "__main__":
    set_dummy = [{"name":"Table","object_id":1,"coord":{"x":3.8130001608398256,"y":-6.793348368367841,"z":0.72},"quaternion":{"x":-0.0,"y":-0.0,"z":-0.624888611604888,"w":0.7807139188502504}},{"name":"Table","object_id":2,"coord":{"x":3.764661804893422,"y":-4.8075594909142705,"z":0.72},"quaternion":{"x":-0.0,"y":-0.0,"z":-0.6510776576139041,"w":0.7590111222874086}},{"name":"Table","object_id":3,"coord":{"x":5.923418155603212,"y":-4.7711079094827955,"z":0.72},"quaternion":{"x":-0.0,"y":-0.0,"z":-0.5979416142212324,"w":0.8015396596441794}},{"name":"Table","object_id":4,"coord":{"x":6.229717575731127,"y":-6.968666381007683,"z":0.72},"quaternion":{"x":-0.0,"y":-0.0,"z":-0.2635740445654651,"w":0.9646391672700224}},{"name":"Chair","object_id":1,"coord":{"x":3.4693877551020407,"y":-7.142857142857142,"z":1.099999999999996},"quaternion":{"x":-0.0,"y":-0.0,"z":-0.0,"w":1.0}},{"name":"Chair","object_id":2,"coord":{"x":3.4693877551020407,"y":-5.1020408163265305,"z":1.099999999999996},"quaternion":{"x":-0.0,"y":-0.0,"z":-0.7210239105529052,"w":0.6929101820661869}},{"name":"Chair","object_id":8,"coord":{"x":3.8775510204081627,"y":-5.5102040816326525,"z":1.099999999999996},"quaternion":{"x":-0.0,"y":-0.0,"z":-0.0,"w":1.0}},{"name":"Chair","object_id":3,"coord":{"x":4.2857142857142865,"y":-4.6938775510204085,"z":1.099999999999996},"quaternion":{"x":-0.0,"y":-0.0,"z":-0.6658136181659445,"w":0.746118104501408}},{"name":"Chair","object_id":4,"coord":{"x":4.6938775510204085,"y":-7.142857142857142,"z":1.099999999999996},"quaternion":{"x":-0.0,"y":-0.0,"z":-0.0,"w":1.0}},{"name":"Chair","object_id":5,"coord":{"x":5.1020408163265305,"y":-5.1020408163265305,"z":1.099999999999996},"quaternion":{"x":-0.0,"y":-0.0,"z":-0.0,"w":1.0}},{"name":"Chair","object_id":6,"coord":{"x":5.510204081632654,"y":-7.142857142857142,"z":1.099999999999996},"quaternion":{"x":-0.0,"y":-0.0,"z":-0.0,"w":1.0}}]
    res = CreateYaml.update(set_dummy, "./cnoid_item_list1.yaml", "./cnoid_item_list2.yaml")
    # pass
    # main_yaml_data = CreateYaml.read("./cnoid_item_base_list.yaml")
    # main_yaml_data = CreateYaml.del_element(main_yaml_data)
    # CreateYaml.write(main_yaml_data, "./cnoid_item_list1.yaml")
