import os
import datetime
import logging
import json
import queue
import time

from mqtt_interface import MqttInterface
from create_yaml import CreateYaml, DELETE_MODEL_NAME_TBL, DYNAMIC_OBJECT_MODEL_TBL
from rearrangement_obstacles_4_lictia import *


FORMAT = "%(asctime)s-%(levelname)s: %(message)s :fn[%(module)s:%(funcName)s]"
LOG_TIME = datetime.datetime.now(datetime.timezone(datetime.timedelta(hours=9), "JST"))
LOG_DIR_PATH = os.path.join(os.path.dirname(__file__), "./log")
LOG_FILE = LOG_DIR_PATH + "/log_" + LOG_TIME.strftime("%Y%m%d%H%M%S") + ".log"

if not os.path.exists(LOG_DIR_PATH):
    os.mkdir(LOG_DIR_PATH)

logger = logging.getLogger("VIEWER")
logger.setLevel(logging.DEBUG)
handler = logging.FileHandler(LOG_FILE)
handler.setLevel(logging.DEBUG)
formatter = logging.Formatter(FORMAT)
handler.setFormatter(formatter)
logger.addHandler(handler)

MQTT_TOPIC = "data/object/EstimatedList"
MQTT_CONF_PATH = os.path.join(os.path.dirname(__file__), "./mqtt_conf.json")
BASE_YAML_FILE_PATH = os.path.join(os.path.dirname(__file__), "./cnoid_item_base_list.yaml")
VIEW_YAML_FILE_PATH = os.path.join(os.path.dirname(__file__), "./cnoid_item_list.yaml")


class ObjectViewer:

    def __init__(self):
        self.mqtt_ins = None
        self.mqtt_queue = queue.Queue()

    def setting(self):
        """
        Connection settings

        :return: Setting result
        """
        result = False
        try:
            with open(MQTT_CONF_PATH, "r") as conf_file:
                mqtt_config = json.load(conf_file)
                self.mqtt_ins = MqttInterface(mqtt_config["host"], mqtt_config["port"], sub_topic=MQTT_TOPIC, mqtt_queue=self.mqtt_queue)
                if self.mqtt_ins.connection(mqtt_config["ca_certs"], mqtt_config["cert_file"], mqtt_config["key_file"]):
                    result = True
        except json.JSONDecodeError as err:
            logger.error("Unable to decode json file")
        except Exception as err:
            logger.error("Unable to read setting file")
        return result

    def run(self, time_up):
        """
        Main processing
        :param time_up: Execution time (minutes)
        :return:
        """
        try:
            end_time = datetime.datetime.now() + datetime.timedelta(minutes=time_up)
            base_file = BASE_YAML_FILE_PATH
            view_file = VIEW_YAML_FILE_PATH
            view_object_a = DELETE_MODEL_NAME_TBL
            if world_item.childItem is None:
                sceneViewProperties()
                # View function call
                readYaml(base_file)
                # Generate world item.
                worldItemGenerator()
                # Generate simulator item
                simulatorItemGenerator(item_list, "Simulator")
                # Read robot models.
                readRobotModel()
                # Read environment models.
                set_folder = readEnvironmentModel()
            else:
                set_folder = env_folder

            semi_static_object_removal_time = None
            while datetime.datetime.now() < end_time:
                if not self.mqtt_queue.empty():
                    logger.info("mqtt_queue data exists")
                    current_time = datetime.datetime.now()
                    if semi_static_object_removal_time is None or semi_static_object_removal_time < current_time:
                        view_object_a = DELETE_MODEL_NAME_TBL
                        semi_static_object_removal_time = current_time + datetime.timedelta(seconds=20)
                    else:
                        view_object_a = DYNAMIC_OBJECT_MODEL_TBL

                    delete_environment_model(view_object_a)

                    object_data = self.mqtt_queue.get(block=True, timeout=None)
                    object_a = object_data.get("object")
                    if object_a is None:
                        self.mqtt_queue.task_done()
                        continue
                    
                    if not CreateYaml.update(object_a, base_file, view_file):
                        logger.error("Unable to update YAML file")
                        self.mqtt_queue.task_done()
                        continue

                    item_a = CreateYaml.read(view_file)
                    set_folder = set_body_model(item_a, env_item, view_object_a, set_folder, "Environment")
                    self.mqtt_queue.task_done()
                time.sleep(0.5)
        except KeyboardInterrupt as err:
            logger.info(err)
        except Exception as err:
            logger.error(err)
        finally:
            self.mqtt_ins.disconnect()


if __name__ == "__main__":
    ins = ObjectViewer()
    if ins.setting():
        ins.run(3)

