/*
 * OpenManipulatorRosController.cpp
 *
 *  Created on: 2020/09/10
 *      Author: Tsuyoshi Anazawa
 *
 *     Outline: RosOpen Manipulatorの制御コントローラ
 *
 */
#include <cnoid/SimpleController>
#include <cnoid/Joystick>
#include <cnoid/EigenUtil>
#include <fmt/format.h>
#include <ros/node_handle.h>
#include <sensor_msgs/Joy.h>
#include <sensor_msgs/JointState.h>
#include <mutex>
#include <iostream>
#include <math.h>

#define rad_to_deg(deg) (deg * M_PI / 180)

using namespace std;
using namespace cnoid;
using fmt::format;

class OpenManipulatorRosController : public cnoid::SimpleController
{
	ros::NodeHandle node;
	ros::Subscriber joystickSubscriber;
	sensor_msgs::Joy latestJoystickState;
	std::mutex joystickMutex;
	ros::Subscriber joint_states_sub_;

	static const int JOINT_NUM = 6;		// Number of joints
	std::vector<Link*> jointArm;		// Store joint links
									   /* First joint,	Second joint,	Third joint */
	const std::string jointName[JOINT_NUM] = { "Joint2",		"Joint3",		"Joint4",
									   /* Fourth joint,	Left gripper,	Right gripper */
			                              "Joint5", 	"Gripper_left", "Gripper_right" };
//	SharedJoystickPtr joystick;
//	int targetMode;
	cnoid::Joystick joystick;						// Variable to store joystick info
	cnoid::Body* body;								// Variable to store Body info
	cnoid::Link::ActuationMode actuationMode;		// Store actuation mode
	double dt;								// Time step
	double q_ref[JOINT_NUM];				// Current angles
	double q_prev[JOINT_NUM];				// Previous angles
	double deltaq;							// Amount of change in joint angle
	double torque;							// torque
	const double MAX_TORQUE = 4.8;			// Max torque
	const double MAX_GRIPPER_TORQUE = 3.0;			// Max torque
	const double ARM_ANGLE_RANGE[5][2] = { /* MIN,					MAX */
										 {	  rad_to_deg( -162.0 ),	rad_to_deg( 162.0 ) },
										 {	  rad_to_deg( -102.6 ),	rad_to_deg(  90.0 ) },
										 {	  rad_to_deg(  -54.0 ),	rad_to_deg(  79.2 ) },
										 {	  rad_to_deg( -102.6 ),	rad_to_deg( 117.0 ) },
										 {					 -0.01,				  0.019 } };

public:
	virtual bool configure(cnoid::SimpleControllerConfig* config) override
	{
		return true;
	}

	virtual bool initialize(cnoid::SimpleControllerIO* io) override
	{
		std::ostream& os = io->os();
		// get body
		body = io->body();

		// Initialize the actuation mode
		actuationMode = cnoid::Link::JOINT_TORQUE;
		// Get controller options
		std::string option = io->optionString();

		if(!option.empty()){
			// Case of the controller option is empty
			if(option == "velocity"){
				// Case of velocity
				actuationMode = cnoid::Link::JOINT_VELOCITY;
			} else if(option == "position"){
				// Case of position
				actuationMode = cnoid::Link::JOINT_DISPLACEMENT;
			} else if(option == "torque" || option == "effort"){
				// Case of torque
				actuationMode = cnoid::Link::JOINT_EFFORT;
			} else {
				// In other case
				os << format("Warning unknown option \"{}\".", option) << std::endl;
			}
		}

		for(int i = 0; i < JOINT_NUM; ++i){
			// Get links
			jointArm.push_back(body->link(jointName[i]));

			if(!jointArm[i]){
				// Error if the link does not exist
				printErrorMessage(os, jointName[i]);
				return false;
			}

			// Set actuation mode for all joints
			jointArm[i]->setActuationMode(actuationMode);
			// Enable joint io
			io->enableIO(jointArm[i]);
			// Set initial joint angle
			q_ref[i] = q_prev[i] = jointArm[i]->q();
		}

		joystickSubscriber = node.subscribe("joy", 1, &OpenManipulatorRosController::joystickCallback, this);
		joint_states_sub_ = node.subscribe("joint_states", 1, &OpenManipulatorRosController::jointStatesCallback, this);
		// Get time step
		dt = io->timeStep();
//		joystick = io->getOrCreateSharedObject<SharedJoystick>("joystick");
//		targetMode = joystick->addMode();

		return true;
	}

	void joystickCallback(const sensor_msgs::Joy& msg)
	{
		std::lock_guard<std::mutex> lock(joystickMutex);
		latestJoystickState = msg;
	}

	void jointStatesCallback(const sensor_msgs::JointState::ConstPtr& msg)
	{
		for(int i = 0; i < jointArm.size(); ++i){
			jointArm[i]->q_target() = msg->position[i];
			q_ref[i] = msg->position[i];
		}
	}

	virtual bool control() override
	{
		sensor_msgs::Joy joystick;
		{
			std::lock_guard<std::mutex> lock(mutex);
			joystick = latestJoystickState;
			joystick.axes.resize(10, 0.0f);
			joystick.buttons.resize(10.0, 0);
		}

		// PD control values
//		static const double Kp[] = { 20.0, 20.0, 20.0, 20.0, 1000.0, 1000.0 };
//		static const double Kd[] = {  2.0,  0.3,  0.3,  0.3, 80.0, 80.0 };
		static const double Kp[] = { 20.0, 15.0, 8.0, 8.0, 1000.0, 1000.0 };
		static const double Kd[] = {  2.0,  1.0,  0.3,  0.3, 80.0, 80.0 };

		// Joystick ID
//		static const int axisID[] = { 4, 5, 3, 2 };
		static const int axisID[] = { 4, 5, 3, 2 };
		double pos[4];

		// Get joystick state
//		joystick->updateState(targetMode);

		// Control arm
		for(int i = 0; i < JOINT_NUM - 2; ++i){
			// Set the current angle
			double q = jointArm[i]->q();
			double q_old;
			// Set the target angular velocity
			double dq_ref = 0.0;
			// Get the amount of change in the joystick
//			double pos = -joystick->getPosition(targetMode, axisID[i], 0.25);
			pos[i] = joystick.axes[axisID[i]];

			if(fabs(pos[i]) > 0.25){
				// Calculate the amount of change in joint angle
				setDeltaq(pos[i], 2.5);

				if(q < ARM_ANGLE_RANGE[i][0]){
					// 現在角が最小関節角より小さい場合
					if(pos[i] < 0){
						deltaq = -deltaq;
						q = ARM_ANGLE_RANGE[i][0];
					}
				} else if(q > ARM_ANGLE_RANGE[i][1]){
					// 現在角が最大関節角より大きい場合
					if(pos[i] > 0){
						deltaq = -deltaq;
						q = ARM_ANGLE_RANGE[i][1];
					}
				}

				// 目標関節角度を算出
				q_ref[i] += deltaq;
				// 目標関節角速度を算出
				dq_ref = deltaq / dt;
			}

			dq_ref = 0.01;
			// 現在の角速度を算出
			double dq = (q - q_prev[i]) / dt;
			// トルクを算出
			torque = Kp[i] * (q_ref[i] - q) + Kd[i] * (dq_ref - dq);
//			torque = Kp[i] * (q_ref[i] - q) + Kd[i] * (0.0 - dq);

			checkTorque(MAX_TORQUE);

			//  関節トルクに算出したトルクを設定
			jointArm[i]->u() = torque;
			// 前回の関節角に現在の関節角を設定
			q_prev[i] = q;
		}

		// 指の制御
		for(int i = 4; i < JOINT_NUM; ++i){
			double q = jointArm[i]->q();
			double dq_ref = 0.0;
			double q_old;
//			double open = joystick->getButtonState(targetMode, 1);
//			double close = -joystick->getButtonState(targetMode, 2);
			double open = joystick.buttons[1];
			double close = -joystick.buttons[2];

			if(close || open){
//				ROS_INFO("q - q_old: %lf, q: %lf, q_old: %lf, deltaq: %lf", q - q_old, q, q_old, deltaq);
				if(fabs(open) > 0.25){
					setDeltaq(open, 30);

					if(q > ARM_ANGLE_RANGE[4][1]){
						deltaq = -deltaq;
						q = ARM_ANGLE_RANGE[4][1];
					}
					q_old = q;
				} else if(fabs(close) > 0.25){
					setDeltaq(close, 30);

					if(q < ARM_ANGLE_RANGE[4][0]){
						deltaq = -deltaq;
						q = ARM_ANGLE_RANGE[4][0];
					}
					q_old = q;
				}
				q_ref[i] += deltaq;
				if(q_ref[i] > ARM_ANGLE_RANGE[4][1]){
					q_ref[i] = ARM_ANGLE_RANGE[4][1];
				} else if(q_ref[i] < ARM_ANGLE_RANGE[4][0]){
					q_ref[i] = ARM_ANGLE_RANGE[4][0];
				}
				dq_ref = deltaq / dt;
			}

//			q_ref[i] += deltaq;
//			dq_ref = deltaq / dt;
//					if(q < ARM_ANGLE_RANGE[4][0] || q - q_prev[i] == 0.0){
//					if(q < ARM_ANGLE_RANGE[4][0]){
//						deltaq = -deltaq;
//						q = ARM_ANGLE_RANGE[4][0];
//					} else if(q > ARM_ANGLE_RANGE[4][1]){
//						q = ARM_ANGLE_RANGE[4][1];
//						q_ref[i] = q;
//						dq_ref = 0.0;
//					} else {
//						q_ref[i] += deltaq;
//						dq_ref = deltaq / dt;
//					}

			double dq = (q - q_prev[i]) / dt;
			torque = Kp[i] * (q_ref[i] - q) + Kd[i] * (dq_ref - dq);
//			ROS_INFO("q_ref: %lf, q: %lf, q_prev: %lf, q - q_prev: %lf, deltaq: %lf, dq_ref: %lf, dq: %lf", q_ref[i], q, q_prev[i], q - q_prev[i], deltaq, dq_ref, dq);
			checkTorque(MAX_GRIPPER_TORQUE);
//			ROS_INFO("i: %d, torque: %lf", i, torque);
			jointArm[i]->u() = torque;
			q_prev[i] = q;
		}

		return true;
	}

	void printErrorMessage(std::ostream& os, std::string name)
	{
		os << "The " << name << " link is not found." << std::endl;
	}

	void setDeltaq(double pos, double divisor)
	{
		deltaq = 0.001 * pos / divisor;
	}

	void checkTorque(double max_torque)
	{
		if(torque > max_torque){
			// トルクが最大値より大きい場合
			torque = max_torque;
		} else if(torque < -max_torque){
			// トルクが最小値より小さい場合
			torque = -max_torque;
		} else{
			// それ以外の場合
		}
	}
};

CNOID_IMPLEMENT_SIMPLE_CONTROLLER_FACTORY(OpenManipulatorRosController)
