/*
 * TurtleBot3RosController.cpp
 *
 *  Created on: 2020/03/25
 *      Author: Tsuyoshi Anazawa
 *
 *  Outline: TurtleBot3モデルをPS4コントローラを用いて動作させるプログラム
 */

#include <cnoid/SimpleController>
#include <fmt/format.h>
#include <ros/node_handle.h>
#include <mutex>
#include <iostream>
#include <ros/ros.h>
#include <nav_msgs/Odometry.h>
#include <tf/transform_broadcaster.h>

using namespace std;
using namespace cnoid;
using fmt::format;

class TFBroadcasterController : public SimpleController
{
	ros::NodeHandle nh_;
	// odometryのPublisher
//	ros::Publisher odomPublisher;
	tf::TransformBroadcaster tf_br_;
	ros::Publisher odom_pub_;

	// ホイール数格納定数
	static const int WHEEL_NUM = 2;
	// bodyファイルに定義されたホイール名格納配列
	const string wheelNames[WHEEL_NUM] = { "Left_wheel", "Right_wheel" };
	// ホイール格納配列
	Link* wheels[2];
	double dt;
	SimpleControllerIO* io;

	double update_period, currentTime, lastTime;
	double update_rate = 3.0;
	double x, y, th;
	Vector2 prevWheelTh;
	double prevVth;
	double d = 0.1435;
	double r = 0.033;

	// ROSトピック配信時のフレームID
	const string ODOM_FRAME_ID = "tb3_0/odom";
	const string CHILD_FRAME_ID = "tb3_0/base_footprint";
	const string IMU_FRAME_ID = "tb3_0/base_footprint";

public:

	virtual bool configure(SimpleControllerConfig* config) override
	{
		return true;
	}

	virtual bool initialize(SimpleControllerIO* io) override
	{
		this->io = io;
		ostream& os = io->os();
		BodyPtr body = io->body();

		for(int i = 0; i < WHEEL_NUM; ++i){
			wheels[i] = body->link(wheelNames[i]);
			if(!wheels[i]){
				os << format("{0} of {1} is not found.", wheelNames[i], body->name()) << endl;
				return false;
			}
			wheels[i]->setActuationMode(Link::JointTorque);
			io->enableInput(wheels[i]);
		}

		for(int i = 0; i < sizeof(wheels) / sizeof(wheels[0]); ++i){
			prevWheelTh[i] = wheels[i]->q();
		}

		odom_pub_ = nh_.advertise<nav_msgs::Odometry>("odom", 10);

		x = -0.986;
		y = -1.900;
		th = 1.57;
		prevVth = 0.0;

		currentTime = lastTime = io->currentTime();
		update_period = 1.0 / update_rate;

		return true;
	}

	void updateOdom()
	{
		currentTime = io->currentTime();
		double updateSince = currentTime - lastTime;

		if(updateSince > update_period){
			currentTime = io->currentTime();
			double deltaT = (currentTime - lastTime);
			Vector2 wheel_th, wheel_omg, wheel_v;
			for(int i = 0; i < sizeof(wheels) / sizeof(wheels[0]); ++i){
				wheel_th[i] = (wheels[i]->q() - prevWheelTh[i]);
				wheel_omg[i] = wheel_th[i] / update_period;
				wheel_v[i] = r * wheel_omg[i];
				prevWheelTh[i] = wheels[i]->q();
			}

			double vx =  (wheel_v[0] + wheel_v[1]) / 2.0;
			double vy = 0.0;
			double omega = (wheel_v[1] - wheel_v[0]) / (2.0 * d);
			double deltaX = (vx * cos(th) - vy * sin(th)) * deltaT;
			double deltaY = (vx * sin(th) + vy * cos(th)) * deltaT;
			double deltaTh = ((omega + prevVth) / 2.0)  * deltaT;

			x += deltaX;
			y += deltaY;
			th += deltaTh;
			// ヨーの値をクオータニオンに変換
			geometry_msgs::Quaternion odomQuat = tf::createQuaternionMsgFromYaw(th);
			// tfに送る TransformedStamped メッセージを作成
			geometry_msgs::TransformStamped odomTrans;
			odomTrans.header.stamp.fromSec(currentTime);
			odomTrans.header.frame_id = ODOM_FRAME_ID;
			odomTrans.child_frame_id = CHILD_FRAME_ID;

			// オドメトリデータから座標変換のデータを代入し, TransformBroadcasterを使用して送信
			odomTrans.transform.translation.x = x;
			odomTrans.transform.translation.y = y;
			odomTrans.transform.translation.z = 0.0;
			odomTrans.transform.rotation = odomQuat;
			tf_br_.sendTransform(odomTrans);

			// ナビゲーションスタックが速度情報を得るために, nav_msgs/Odometry メッセージを配信
			nav_msgs::Odometry odom;
			odom.header.stamp.fromSec(currentTime);
			odom.header.frame_id = ODOM_FRAME_ID;

			odom.pose.pose.position.x = x;
			odom.pose.pose.position.y = y;
			odom.pose.pose.position.z = 0.0;
			odom.pose.pose.orientation = odomQuat;

			odom.child_frame_id = CHILD_FRAME_ID;
			odom.twist.twist.linear.x = vx;
			odom.twist.twist.linear.y = vy;
//			odom.twist.twist.angular.z = vth[2];
			odom.twist.twist.angular.z = omega;
//			odom.twist.twist.linear.x = cmdvel[0];
//			odom.twist.twist.linear.y = cmdvel[1];
//			odom.twist.twist.angular.z = cmdvelTh[2];

			ros::spinOnce();
			odom_pub_.publish(odom);

			lastTime += update_period;
			prevVth = omega;
		}
	}

	virtual bool control() override
	{
		updateOdom();
		return true;
	}

	virtual void stop() override
	{
	}
};

CNOID_IMPLEMENT_SIMPLE_CONTROLLER_FACTORY(TFBroadcasterController)
