/*
 * TurtleBot3ReadCoordController.cpp
 *
 *  Created on: 2022/01/19
 *      Author: Tsuyoshi Anazawa
 *
 *  Outline: Read coordinates and control TurtleBot3.
 */

#include <cnoid/SimpleController>
#include <cnoid/EigenUtil>
#include <fmt/format.h>
#include <fstream>
#include <iostream>
#include <math.h>
#include <cmath>
#include <tf/transform_broadcaster.h>
#include <Eigen/Geometry>

using namespace std;
using namespace cnoid;
using fmt::format;

class TurtleBot3ReadCoordController : public SimpleController
{
	SimpleControllerIO* io;
	Body* body;
	Link* rootLink;
	Vector3 T;
	Matrix3 R;
	double dt;

	// The input file operations.
	string ext = ".csv";
	string fileName = "odom" + ext;
	// Array definition
	vector<double> ift, ifx, ify, ifz, ifori_x, ifori_y, ifori_z, ifori_w,
	ifroll, ifpitch, ifyaw;
	// CSV data storage variables.
	string tmp;
	// File I/O declaration.
	FILE* fp;
	// Buffer size.
	static const int BUF_SIZE = 100;
	char str[BUF_SIZE];
	// Command to get the current directory.
	string cmd = "pwd | tr '\n' '/'";
	int cnt = 0;
	// File path.
	string filePath = "";
	int id = 0;
	bool first_line = true;
	int data_num = 0;
	double rate = 100.0;
	double ms = 0.0;
	double dx, dy, dz, droll, dpitch, dyaw;
	// Output csv file.
	ofstream ofs;
	double last_time;
	double sec;
	double first_time;

public:
	virtual bool initialize(SimpleControllerIO* io) override
	{
		this->io = io;
		ostream& os = io->os();
		// Body variable.
		body = io->body();

		// Get controller option.
		string option = io->optionString();

		// Store root link in array.
		rootLink = body->rootLink();
		rootLink->setActuationMode(LINK_POSITION);
		if(!rootLink){
			// When there is no root link.
			os << format("{0} of {1} is not found.", rootLink->name(), body->name()) << endl;
			return false;
		}

		// Enable input to root link.
		io->enableIO(rootLink);

		dt = io->timeStep();

		if((fp = popen(cmd.c_str(), "r")) != NULL){
			// Open a process and execute a command.
			while(fgets(str, sizeof(str), fp) != NULL){
				// Read command results line by line.
				// Get current directory.
				filePath += str;
			}
			// Close process.
			pclose(fp);
		}

		// Combine directory name and file name.
		filePath = filePath + fileName;
		// Read the file.
//		os << filePath << endl;
		ifstream ifs(filePath);
		if(ifs.fail()){
			cerr << "Failed to open file." << endl;
			cerr << filePath << endl;
			filePath.clear();
			return false;
		}

		while(getline(ifs, tmp)){
			if(ifs.eof()){
				ifs.close();
				break;
			} else if(first_line){
				first_line = false;
				continue;
			} else {
				std::stringstream ss(tmp);
				while(getline(ss, tmp, ',')){
					if(cnt == 5){
						sec = stod(tmp);
						cnt++;
					} else if(cnt == 6){
						double time = sec + stod(tmp) / pow(10, 9);
						if(ift.size() == 0){
							first_time = time;
						}
						time -= first_time;
						ift.push_back(time);
						cnt++;
					} else if(cnt == 8){
						ifx.push_back(stod(tmp));
						cnt++;
					} else if(cnt == 9){
						ify.push_back(stod(tmp));
						cnt++;
					} else if(cnt == 10){
						ifz.push_back(stod(tmp));
						cnt++;
					} else if(cnt == 11){
						ifori_x.push_back(stod(tmp));
						cnt++;
					} else if(cnt == 12){
						ifori_y.push_back(stod(tmp));
						cnt++;
					} else if(cnt == 13){
						ifori_z.push_back(stod(tmp));
						cnt++;
					} else if(cnt == 14){
						ifori_w.push_back(stod(tmp));
						cnt = 0;
					} else {
						cnt++;
					}
				}
			}
		}

		quat_to_RPY();

		string file_path = "/home/fsk9001/.ros/cnoid_coord_readcoord.csv";
		ifstream out_ifs(file_path);
		string hdr;
		if(out_ifs.fail()){
			hdr = "sim_time,x,y,z";
		}
		ofs.open(file_path, ios::app);
		if(!hdr.empty()){
			ofs << hdr << endl;
		}

		last_time = io->currentTime();
		dt = io->timeStep();

		id = 0;
//		ms = 1.0 / rate;
		ms = dt;

		return true;
	}

	void quat_to_RPY()
	{
		for(int i = 0; i < ifori_x.size(); ++i){
			tf::Quaternion quat;
			double roll, pitch, yaw;
			quat.setValue(ifori_x[i], ifori_y[i], ifori_z[i], ifori_w[i]);
			tf::Matrix3x3(quat).getRPY(roll, pitch, yaw);
			ifroll.push_back(roll);
			ifpitch.push_back(pitch);
			ifyaw.push_back(yaw);
		}
	}

	virtual bool control() override
	{
		double current_time = io->currentTime();
//		double time = ift[id] - ift[0];
		if(id < ift.size() && (io->currentTime() >= ift[id] && io->currentTime() < ift[id] + ms)){
			T = Vector3(ifx[id], ify[id], ifz[id]);
			R = rotFromRpy(ifroll[id], ifpitch[id], ifyaw[id]);
			T += R * Vector3(0.0, 0, 0.015);

			if((id + 1) < ift.size()){
//				dx = (ifx[id + 1] - ifx[id]) / rate;
//				dy = (ify[id + 1] - ify[id]) / rate;
//				dz = (ifz[id + 1] - ifz[id]) / rate;
//				droll = (ifroll[id + 1] - ifroll[id]) / rate;
//				dpitch = (ifpitch[id + 1] - ifpitch[id]) / rate;
//				dyaw = (ifyaw[id + 1] - ifyaw[id]) / rate;
				dx = (ifx[id + 1] - ifx[id]) * dt;
				dy = (ify[id + 1] - ify[id]) * dt;
				dz = (ifz[id + 1] - ifz[id]) * dt;
				droll = (ifroll[id + 1] - ifroll[id]) * dt;
				dpitch = (ifpitch[id + 1] - ifpitch[id]) * dt;
				if((ifyaw[id + 1] <= PI && ifyaw[id + 1] >= PI/2)
						&& (ifyaw[id] >= -PI && ifyaw[id] <= -PI/2)){
					dyaw = -((PI - ifyaw[id + 1]) + (PI + ifyaw[id])) * dt;
				} else if((ifyaw[id + 1] >= -PI && ifyaw[id + 1] <= -PI/2)
						&& (ifyaw[id] >= PI/2 && ifyaw[id] <= PI)){
					dyaw = ((PI + ifyaw[id + 1]) + (PI - ifyaw[id])) * dt;
				} else {
					dyaw = (ifyaw[id + 1] - ifyaw[id]) * dt;
				}
			}

			id++;
			ms = 0.0;
		} else {
			if((id >= 1 && id < ift.size()) && io->currentTime() >= ift[id - 1] + ms){
				ifx[id - 1] += dx;
				ify[id - 1] += dy;
				ifz[id - 1] += dz;
				ifroll[id - 1] += droll;
				ifpitch[id - 1] += dpitch;
				if(ifyaw[id - 1] > 0 && (ifyaw[id - 1] + dyaw) > PI){
					ifyaw[id - 1] = -PI + (ifyaw[id - 1] + dyaw - PI);
				}else if(ifyaw[id - 1] < 0 && (ifyaw[id - 1] + dyaw) < -PI){
					ifyaw[id - 1] = PI + (ifyaw[id - 1] + dyaw - PI);
				} else {
					ifyaw[id - 1] += dyaw;
				}

				T = Vector3(ifx[id - 1], ify[id - 1], ifz[id - 1]);
				R = rotFromRpy(ifroll[id - 1], ifpitch[id - 1], ifyaw[id - 1]);
				T += R * Vector3(0.0, 0.0, 0.015);

//				ms += 1.0 / rate;
			}
			ms += dt;
		}

//		cout << ift[id] << "\t" << ifx[id] << "\t" << ify[id] << "\t" << ifz[id] << endl;
		rootLink->setTranslation(T);
		rootLink->setRotation(R);
		body->calcForwardKinematics();
//
		Isometry3 position = rootLink->position();
		Vector3 rpy = rpyFromRot(position.rotation());

		if(current_time - last_time >= 1.0){
			ofs << current_time << "," << position.translation().x() << ","  << position.translation().y() << "," << 0.0 << endl;
			last_time = current_time;
		}

		return true;
	}

	virtual void stop() override
	{
//		id = 0;
//		filePath.clear();
//		ift.clear();
//		ift.shrink_to_fit();
//		ifx.clear(), ify.clear(), ifz.clear();
//		ifx.shrink_to_fit(), ify.shrink_to_fit(), ifz.shrink_to_fit();
//		ifori_x.clear(), ifori_y.clear(), ifori_z.clear(), ifori_w.clear();
//		ifori_x.shrink_to_fit(), ifori_y.shrink_to_fit(), ifori_z.shrink_to_fit(), ifori_w.shrink_to_fit();
//		ifroll.clear(), ifpitch.clear(), ifyaw.clear();
//		ifroll.shrink_to_fit(), ifpitch.shrink_to_fit(), ifyaw.shrink_to_fit();
//		ms = 0.0;
//		ofs.close();
//		free(io);
//		io = NULL;
//		free(body);
//		body = NULL;
//		free(rootLink);
//		rootLink = NULL;
	}

};

CNOID_IMPLEMENT_SIMPLE_CONTROLLER_FACTORY(TurtleBot3ReadCoordController)
