/*
 * TurtleBot3ReadCoordController.cpp
 *
 *  Created on: 2021/12/02
 *      Author: Tsuyoshi Anazawa
 *
 *      Outline: Read coordinates and control TurtleBot3.
 */

#include <cnoid/SimpleController>
#include <fmt/format.h>
#include <ros/node_handle.h>
#include <iostream>
#include <ros/ros.h>
#include <tf/transform_broadcaster.h>
// simulation time
#include <std_msgs/String.h>
#include <cnoid/EigenUtil>
#include <math.h>
#include <fstream>
#include <cmath>

using namespace std;
using namespace cnoid;
using fmt::format;

class TurtleBot3ReadCoordController : public SimpleController
{
	ros::NodeHandle nh_;

	// Number of wheels
	static const int WHEEL_NUM = 2;
	// Wheel names
	const string wheelNames[WHEEL_NUM] = { "Left_wheel", "Right_wheel" };
	// Actuation mode
	int actuationMode;
	// Wheel links
	Link* wheels[2];
	double qprev[2];
	double dt;
	double qref[2];
	SimpleControllerIO* io;
	Vector3 cmd_vel, cmd_vel_th;
	// Initial position of the robot
	const double INITIAL_POSE_THETA = 0.0;
	// Limit velocities and accelerations
	const double MAX_VEL = 0.26;
	const double MIN_VEL = -0.26;
	const double MAX_ACCEL = 1.82;
	const double MIN_ACCEL = -1.82;
	// half of the tread
	double d = 0.1435;
	// Wheel radius
	double r = 0.033;

	// Simulation time.
	ros::Publisher simtime_pub_;

	// The input file operations.
	string fileName = "odom";
	string ext = ".csv";
	string num;
	// Array definition(t: time, x: x axis, y: y axis).
	vector<double> ift, ifx, ify, ifz, ifori_x, ifori_y, ifori_z, ifori_w,
	ifroll, ifpitch, ifyaw, ifsim_t;
	vector<double> ifvx, ifvth;
	// CSV data storage variables.
	string tmp;
	// File I/O declaration.
	FILE* fp;
	// Buffer size.
	static const int BUF_SIZE = 100;
	char str[BUF_SIZE];
	// Command to get the current directory.
	string cmd = "pwd | tr '\n' '/'";
	int cnt = 0;
	// File path
	string filePath = "";
	int id = 0;
	bool first_line = true;
	double prev_time;
	int data_num = 0;
	// Output csv file.
	BodyPtr ioBody;
	ofstream ofs;
	double last_time;
	double sec;

public:
	double clamp(const double x, const double min, const double max)
	{
	  return std::min(std::max(min, x), max);
	}

	virtual bool configure(SimpleControllerConfig* config) override
	{
		return true;
	}

	virtual bool initialize(SimpleControllerIO* io) override
	{
		this->io = io;
		ostream& os = io->os();
		// Body variable.
		Body* body = io->body();
		ioBody = body;

		// Set joint's actuation mode.
		actuationMode = Link::JointTorque;
		// Get controller option.
		string option = io->optionString();

		if(!option.empty()){
			// When controller option is empty.
			if(option == "velocity" || option == "position"){
				// When controller option is velocity or position.
				actuationMode = Link::JointVelocity;
			} else if(option == "torque"){
				// When controller option is torque.
				actuationMode = Link::JointTorque;
			} else {
				// In other cases.
				os << format("Warning: Unknown option {}.", option) << endl;
			}
		}

		// Root link
		io->enableInput(ioBody->rootLink(), LINK_POSITION);

		for(int i = 0; i < WHEEL_NUM; ++i){
			// Store wheel links in array.
			wheels[i] = body->link(wheelNames[i]);
			if(!wheels[i]){
				// When there is no wheel links.
				os << format("{0} of {1} is not found.", wheelNames[i], body->name()) << endl;
				return false;
			}

			// Set the wheel array's actuation mode.
			wheels[i]->setActuationMode(actuationMode);
			// Enable output to wheel links.
			io->enableIO(wheels[i]);
			// Store current joint angles in array
			qref[i] = qprev[i] = wheels[i]->q();
		}

		// Create publisher for simulation time.
		simtime_pub_ = nh_.advertise<std_msgs::String>("simtime", 10);

		if((fp = popen(cmd.c_str(), "r")) != NULL){
			// Open a process and execute a command.
			while(fgets(str, sizeof(str), fp) != NULL){
				// Read command results line by line.
				// Get current directory.
				filePath += str;
			}
			// Close process.
			pclose(fp);
		}

		// Combine directory name and file name.
		filePath = filePath + "../.ros/" + fileName + ext;
		// Read the file.
		ifstream ifs(filePath);
		if(ifs.fail()){
			cerr << "Failed to open file." << endl;
			filePath.clear();
			return false;
		}

		while(getline(ifs, tmp)){
			if(ifs.eof()){
				ifs.close();
				break;
			} else if(first_line){
				first_line = false;
				continue;
			} else {
				std::stringstream ss(tmp);
				while(getline(ss, tmp, ',')){
//					if(cnt == 1){
//						int basepos = 11, c_num = 2;
//						int hour = stoi(tmp.substr(basepos, c_num));
//						int minute = stoi(tmp.substr(basepos + 3, c_num));
//						int second = stoi(tmp.substr(basepos + 6, c_num));
//						double msecond = stod("0" + tmp.substr(basepos + 8, c_num + 2));
//						double comb_time = hour * 3600 + minute * 60 + second + msecond;
//						ift.push_back(comb_time);
//						cnt++;
//					} else if(cnt == 2){
//						ifsim_t.push_back(stod(tmp));
//						cnt++;
//					} else if(cnt == 3){
//						ifx.push_back(stod(tmp));
//						cnt++;
//					} else if(cnt == 4){
//						ify.push_back(stod(tmp));
//						cnt++;
//					} else if(cnt == 5){
//						ifz.push_back(stod(tmp));
//						cnt++;
//					} else if(cnt == 6){
//						ifori_x.push_back(stod(tmp));
//						cnt++;
//					} else if(cnt == 7){
//						ifori_y.push_back(stod(tmp));
//						cnt++;
//					} else if(cnt == 8){
//						ifori_z.push_back(stod(tmp));
//						cnt++;
//					} else if(cnt == 9){
//						ifori_w.push_back(stod(tmp));
//						cnt = 0;
//					if(cnt == 3){
//						int basepos = 11, c_num = 2;
//						int hour = stoi(tmp.substr(basepos, c_num));
//						int minute = stoi(tmp.substr(basepos + 3, c_num));
//						int second = stoi(tmp.substr(basepos + 6, c_num));
//						double msecond = stod("0" + tmp.substr(basepos + 8, c_num + 2));
//						double comb_time = hour * 3600 + minute * 60 + second + msecond;
//						ift.push_back(comb_time);
//						cnt++;
					if(cnt == 5){
						sec = stod(tmp);
						cnt++;
					} else if(cnt == 6){
						double time = sec + stod(tmp) / pow(10, 9);
						ift.push_back(time);
						cnt++;
					} else if(cnt == 8){
						ifx.push_back(stod(tmp));
						cnt++;
					} else if(cnt == 9){
						ify.push_back(stod(tmp));
						cnt++;
					} else if(cnt == 10){
						ifz.push_back(stod(tmp));
						cnt++;
					} else if(cnt == 11){
						ifori_x.push_back(stod(tmp));
						cnt++;
					} else if(cnt == 12){
						ifori_y.push_back(stod(tmp));
						cnt++;
					} else if(cnt == 13){
						ifori_z.push_back(stod(tmp));
						cnt++;
					} else if(cnt == 14){
						ifori_w.push_back(stod(tmp));
						cnt = 0;
					} else {
						cnt++;
					}
				}
			}
		}

		quat_to_RPY();
		calcVel();

		string file_path = "/home/fsk9001/.ros/cnoid_coord_readcoord.csv";
		ifstream out_ifs(file_path);
		string hdr;
		if(out_ifs.fail()){
			hdr = "sim_time,x,y,z";
		}
		ofs.open(file_path, ios::app);
		if(!hdr.empty()){
			ofs << hdr << endl;
		}

		last_time = io->currentTime();
		dt = io->timeStep();
		prev_time = io->currentTime();

		return true;
	}

	void quat_to_RPY()
	{
		for(int i = 0; i < ifori_x.size(); ++i){
			tf::Quaternion quat;
			double roll, pitch, yaw;
			quat.setValue(ifori_x[i], ifori_y[i], ifori_z[i], ifori_w[i]);
			tf::Matrix3x3(quat).getRPY(roll, pitch, yaw);
			ifroll.push_back(roll);
			ifpitch.push_back(pitch);
			ifyaw.push_back(yaw);
		}
	}

	void calcVel()
	{
		for(int i = 1; i < ifx.size(); ++i){
			double d_t = ift[i] - ift[i - 1];
//			double d_t = ifsim_t[i] - ifsim_t[i - 1];
			double vel_x = sqrt(pow(ifx[i] - ifx[i - 1], 2) + pow(ify[i] - ify[i - 1], 2)) / d_t;
			double vel_th = (ifyaw[i] - ifyaw[i - 1]) / d_t;
			ifvx.push_back(vel_x);
			ifvth.push_back(vel_th);
		}
	}

	double getTime()
	{
		char iso_time[30];
		char dst[50];
		struct timeval ct;
		struct tm* local_time;
		int basepos = 11, c_num = 2;

		memset(iso_time, 0, sizeof(iso_time));
		memset(dst, 0, sizeof(dst));

		gettimeofday(&ct, NULL);
		local_time = localtime(&ct.tv_sec);

		strftime(iso_time, sizeof(iso_time) - 1, "%FT%T", local_time);

		sprintf(dst, "%s.%03lu", iso_time, (ct.tv_usec + 500) / 1000);

		string time_str = dst;

		int hour = stoi(time_str.substr(basepos, c_num));
		int minute = stoi(time_str.substr(basepos + 3, c_num));
		int second = stoi(time_str.substr(basepos + 6, c_num));
		double msecond = stod("0" + time_str.substr(basepos + 8, c_num + 2));
		double comb_time = hour * 3600 + minute * 60 + second + msecond;

		return comb_time;
	}

	virtual bool control() override
	{
		double current_time = io->currentTime();
//		double current_time = getTime();
//		double d_t = ifsim_t[data_num + 1] - ifsim_t[data_num];
		double d_t = ift[data_num + 1] - ift[data_num];
		if((current_time - prev_time) <= d_t){
			cmd_vel[0] = ifvx[data_num];
			cmd_vel_th[2] = ifvth[data_num];
		} else {
			if(data_num > ifx.size()){
				cmd_vel[0] = 0.0;
				cmd_vel_th[2] = 0.0;
			} else {
				prev_time = current_time;
				data_num++;
			}
		}

//		ROS_INFO("cmv_vel.linear: %lf,  cmd_vel.angular: %lf", cmd_vel[0], cmd_vel_th[2]);

		if(actuationMode == Link::JointVelocity){
			// When actuation mode is velocity.
			wheels[0]->dq_target() = 1.17 * clamp(cmd_vel[0] - clamp(cmd_vel_th[2], MIN_ACCEL, MAX_ACCEL) * d, MIN_VEL, MAX_VEL) / r;
			wheels[1]->dq_target() = 1.17 * clamp(cmd_vel[0] + clamp(cmd_vel_th[2], MIN_ACCEL, MAX_ACCEL) * d, MIN_VEL, MAX_VEL) / r;

		} else if(actuationMode == Link::JointTorque){
			// When actuation mode is torque.
			static const double K = 35.0;	// rotorInertia = 1.0e-06
			double dq_target[2];		// Target angular velocity.
			dq_target[0] = clamp(cmd_vel[0] - clamp(cmd_vel_th[2], MIN_ACCEL, MAX_ACCEL) * d, MIN_VEL, MAX_VEL) / r;
			dq_target[1] = clamp(cmd_vel[0] + clamp(cmd_vel_th[2], MIN_ACCEL, MAX_ACCEL) * d, MIN_VEL, MAX_VEL) / r;

			for(int i = 0; i < 2; ++i){
				double q = wheels[i]->q();
				double dq = (q - qprev[i]) / dt;	// Current angular velocity.

				wheels[i]->u() = K * (dq_target[i] - dq);	// Torque value given to tire.
				qprev[i] = wheels[i]->q();	// Update previous joint angular.
			}
		}

		auto T = ioBody->rootLink()->position();
		Vector3 rpy = rpyFromRot(T.rotation());

		if(current_time - last_time >= 1.0){
			ofs << current_time << "," << T.translation().x() << ","  << T.translation().y() << "," << 0.0 << endl;
			last_time = current_time;
		}

		return true;
	}

	virtual void stop() override
	{
		ift.clear(), ifsim_t.clear();
		ifx.clear(), ify.clear(), ifz.clear();
		ifori_x.clear(), ifori_y.clear(), ifori_z.clear(), ifori_w.clear();
		ifroll.clear(), ifpitch.clear(), ifyaw.clear();
		ifvx.clear(), ifvth.clear();
		ofs.close();
	}
};

CNOID_IMPLEMENT_SIMPLE_CONTROLLER_FACTORY(TurtleBot3ReadCoordController)
