/*
 * TurtleBot3RosController.cpp
 *
 *  Created on: 2020/03/25
 *      Author: Tsuyoshi Anazawa
 *
 *  Outline: The controller for operating TurtleBot3 model with PS4 controller.
 */

#include <cnoid/SimpleController>
//#include <cnoid/RangeCamera>
#include <cnoid/RangeSensor>
#include <cnoid/AccelerationSensor>
#include <cnoid/RateGyroSensor>
#include <cnoid/Joystick>
#include <fmt/format.h>
#include <ros/node_handle.h>
#include <sensor_msgs/Joy.h>
#include <sensor_msgs/Imu.h>
#include <mutex>
//#include <iostream>
#include <ros/ros.h>
#include <nav_msgs/Odometry.h>
#include <tf/transform_broadcaster.h>
#include <geometry_msgs/Twist.h>
//#include <geometry_msgs/TransformStamped.h>
// For angle confirmation(Debug).
#include <std_msgs/Float32MultiArray.h>
#include <sensor_msgs/PointCloud.h>
#include <sensor_msgs/PointCloud2.h>
#include <sensor_msgs/point_cloud_conversion.h>

using namespace std;
//using namespace cnoid;
using fmt::format;

class TurtleBot3RosController : public cnoid::SimpleController
{
	ros::NodeHandle nh_;
	// Joystick subscriber
	ros::Subscriber joy_sub_;
	sensor_msgs::Joy latestJoystickState;
	std::mutex joystickMutex;
	// Odometry publisher
	ros::Publisher odom_pub_;
	tf::TransformBroadcaster tfBroadcaster_;
	// 0: odom, 1: imu, 2: cmd_vel
	double currentTime[3], lastTime[3], update_period[3];
	double update_rate[3] = { 3.0, 20.0, 10.0 };
	// hector slamの場合odomを6Hzとしないとエラーが発生
//	double update_rate[3] = { 6.0, 20.0, 10.0 };

	// Imu publisher
	ros::Publisher imu_pub_;
	cnoid::AccelerationSensorPtr accel;
	cnoid::RateGyroSensorPtr gyro;
	std::string deviceName[2] = { "AccelSensor", "GyroSensor" };
	cnoid::ScopedConnection accel_conn;
	cnoid::ScopedConnection gyro_conn;

	// cmd_vel subscriber
	ros::Subscriber cmd_vel_sub_;
	// For angle confirmation(Debug)
	ros::Publisher theta_pub_;
	ros::Publisher pcl2_pub_;
	ros::Subscriber pcl_sub_;
	// Constant to store the number of wheels.
	static const int WHEEL_NUM = 2;
	// The array containing the wheel names defined in the body file.
	const std::string wheelNames[WHEEL_NUM] = { "Left_wheel", "Right_wheel" };
	// The variable that stores the actuation mode of the link.
	int actuationMode;
	// The array to store wheels.
	cnoid::Link* wheels[2];
	double qprev[2];
	double dt;
	double qref[2];
	cnoid::SimpleControllerIO* io;
	// Current coordinate position.
	double x, y, th;
	double omega;
	// The array used in accelerometers and angular velocity sensors.
	// Current velocity[m/s] and angular velocity[rad/s].
	cnoid::Vector3 v;
	// Last velocity[m/s].
	cnoid::Vector3 prevVth;
	double imuTh;
	cnoid::Vector3 cmd_vel, cmd_vel_th;
	const std::string ODOM_FRAME_ID = "odom";
	const std::string CHILD_FRAME_ID = "base_footprint";
	const std::string IMU_FRAME_ID = "base_footprint";
	// Limit velocities and accelerations.
	const double MAX_VEL = 0.26;
	const double MIN_VEL = -0.26;
	const double MAX_ACCEL = 1.82;
	const double MIN_ACCEL = -1.82;
	// half of the tread.
	double d = 0.1435;
	// Wheel radius.
	double r = 0.033;
	cnoid::Vector2 prevWheelTh;
	cnoid::Vector3 dv_sum[2], w_sum[2];
	int sensingSteps[2];

public:

	double clamp(const double x, const double min, const double max)
	{
	  return std::min(std::max(min, x), max);
	}

	virtual bool configure(cnoid::SimpleControllerConfig* config) override
	{
		return true;
	}

	virtual bool initialize(cnoid::SimpleControllerIO* io) override
	{
		this->io = io;
		ostream& os = io->os();
		// Body variable.
		cnoid::Body* body = io->body();

		// Set joint's actuation mode.
		actuationMode = cnoid::Link::JointTorque;
		// Get controller option.
		std::string option = io->optionString();

		if(!option.empty()){
			// When controller option is empty.
			if(option == "velocity" || option == "position"){
				// When controller option is velocity or position.
				actuationMode = cnoid::Link::JointVelocity;
			} else if(option == "torque"){
				// When controller option is torque.
				actuationMode = cnoid::Link::JointTorque;
			} else {
				// In other cases.
				os << format("Warning: Unknown option {}.", option) << endl;
			}
		}

		for(int i = 0; i < WHEEL_NUM; ++i){
			// Store wheel links in array.
			wheels[i] = body->link(wheelNames[i]);
			if(!wheels[i]){
				// When there is no wheel links.
				os << format("{0} of {1} is not found.", wheelNames[i], body->name()) << endl;
				return false;
			}

			// Set the wheel array's actuation mode.
			wheels[i]->setActuationMode(actuationMode);
			// Enable output to wheel links.
			io->enableOutput(wheels[i]);
			// Store current joint angles in array
			qprev[i] = wheels[i]->q();
		}

		accel = body->findDevice<cnoid::AccelerationSensor>(deviceName[0]);
		if(!accel){
			os << format("{0} of {1} is not found.", deviceName[0], body->name()) << endl;
			//return false;
		} else {
			io->enableInput(accel);
			accel_conn.disconnect();
			accel_conn = accel->sigStateChanged().connect(
					[&](){ updateImu(); });
		}
		
		gyro = body->findDevice<cnoid::RateGyroSensor>(deviceName[1]);
		if(!gyro){
			os << format("{0} of {1} is not found.", deviceName[1], body->name()) << endl;
			//return false;
		} else {
			io->enableInput(gyro);
			gyro_conn.disconnect();
			gyro_conn = gyro->sigStateChanged().connect(
					[&](){ updateImu(); });
		}

		// joystick subscriber
		joy_sub_ = nh_.subscribe("joy", 1, &TurtleBot3RosController::joystickCallback, this);
		// cmd_vel subscriber
		cmd_vel_sub_ = nh_.subscribe("waffle_pi/cmd_vel", 1, &TurtleBot3RosController::cmdvelCallback, this);
		// Create the publisher of odometry(topic: odom, buffer_size: 10).
//		odomPublisher = nh_.advertise<nav_msgs::Odometry>("waffle_pi/odom", 10);
		odom_pub_ = nh_.advertise<nav_msgs::Odometry>("odom", 10);
		// Create the publisher of imu(topic: imu, buffer_size: 10).
//		imuPublisher = nh_.advertise<sensor_msgs::Imu>("waffle_pi/imu", 10);
		imu_pub_ = nh_.advertise<sensor_msgs::Imu>("imu", 10);
		// For angle confirmation(Debug).
		theta_pub_ = nh_.advertise<std_msgs::Float32MultiArray>("waffle_pi/theta", 10);
		pcl_sub_ = nh_.subscribe("waffle_pi/velodyne/point_cloud", 1, &TurtleBot3RosController::pointcloudCallback, this);
//		pointcloud2pub_ = nh_.advertise<sensor_msgs::PointCloud2>("velodyne_points", 10);
		pcl2_pub_ = nh_.advertise<sensor_msgs::PointCloud2>("multi_scan_points", 10);

		// Suppose the robot starts from the origin of the odom coordinate system.
		x = 0.0;
		y = 0.0;
		th = 0.0;

		prevVth[2] = 0.0;

		currentTime[3] = lastTime[3] = io->currentTime();
		for(int i = 0; i < sizeof(currentTime) / sizeof(currentTime[0]); ++i){
			update_period[i] = 1.0 / update_rate[i];
		}

		dt = io->timeStep();

		return true;
	}

	void joystickCallback(const sensor_msgs::Joy& msg)
	{
		std::lock_guard<std::mutex> lock(joystickMutex);
		latestJoystickState = msg;
	}

	void cmdvelCallback(const geometry_msgs::Twist& msg)
	{
		cmd_vel[0] = msg.linear.x;
		cmd_vel[1] = msg.linear.y;
		cmd_vel[2] = msg.linear.z;
		cmd_vel_th[0] = msg.angular.x;
		cmd_vel_th[1] = msg.angular.y;
		cmd_vel_th[2] = msg.angular.z;
		ROS_INFO("------------------ Subscribe of cmd_vel ------------------");
		ROS_INFO("linear: ( %lf, %lf, %lf )", msg.linear.x, msg.linear.y, msg.linear.z);
		ROS_INFO("angular: ( %lf, %lf, %lf )", msg.angular.x, msg.angular.y, msg.angular.z);
		ROS_INFO("------------------ Subscribe of cmd_vel ------------------");
	}

	void pointcloudCallback(const sensor_msgs::PointCloud& msg)
	{
		sensor_msgs::PointCloud2 pc2_out;
		sensor_msgs::PointCloud pc_in;
		pc_in.header = msg.header;
		pc_in.channels = msg.channels;
//		ROS_INFO("%d", msg.points.size());
		for(int i = 0; i < msg.points.size(); i++){
//			ROS_INFO("%lf  %lf  %lf", msg.points[i].x, msg.points[i].y, msg.points[i].z);
			geometry_msgs::Point32 point;
			point.x = -msg.points[i].z;
			point.y = -msg.points[i].x;
			point.z = msg.points[i].y;
			pc_in.points.push_back(point);
		}
		sensor_msgs::convertPointCloudToPointCloud2(pc_in, pc2_out);
		pcl2_pub_.publish(pc2_out);
	}

	void updateLocalization()
	{
		cnoid::Vector2 wheel_th, wheel_omg, wheel_v;

		for(int i = 0; i < sizeof(wheels) / sizeof(wheels[0]); ++i){
			double latest_wheel_angle = wheels[i]->q();
			wheel_th[i] = (latest_wheel_angle - prevWheelTh[i]);
			wheel_omg[i] = wheel_th[i] / update_period[0];
			wheel_v[i] = r * wheel_omg[i];
			prevWheelTh[i] = latest_wheel_angle;
		}

		v[0] = clamp((wheel_v[0] + wheel_v[1]) / 2.0, MIN_VEL, MAX_VEL);
		v[1] = 0.0;

		double deltaX = (v[0] * cos(th) - v[1] * sin(th)) * update_period[0];
		double deltaY = (v[0] * sin(th) + v[1] * cos(th)) * update_period[0];
		double deltaTh = ((omega + prevVth[2]) / 2.0) * update_period[0];

		x += deltaX;
		y += deltaY;
		th += deltaTh;

		prevVth[2] = omega;
	}

	void updateOdom()
	{
		currentTime[0] = io->currentTime();
		double updateSince = currentTime[0] - lastTime[0];
		w_sum[0] += gyro->w();
		sensingSteps[0]++;

		if(updateSince > update_period[0]){
			cnoid::Vector3 w = w_sum[0] / sensingSteps[0];
			// Convert yaw values to quaternions.
			geometry_msgs::Quaternion odomQuat = tf::createQuaternionMsgFromYaw(th);
			// Create a TransformedStamped message to send to tf.
			geometry_msgs::TransformStamped odomTrans;
			odomTrans.header.stamp.fromSec(currentTime[0]);
			odomTrans.header.frame_id = ODOM_FRAME_ID;
			odomTrans.child_frame_id = CHILD_FRAME_ID;

			// Substitute coordinate transformation data from odometry data and send using Transform Broadcaster.
			odomTrans.transform.translation.x = x;
			odomTrans.transform.translation.y = y;
			odomTrans.transform.translation.z = 0.0;
			odomTrans.transform.rotation = odomQuat;
			tfBroadcaster_.sendTransform(odomTrans);

			// The navigation stack publishes a nav_msgs/Odometry message for velocity information.
			nav_msgs::Odometry odom;
			odom.header.stamp.fromSec(currentTime[0]);
			odom.header.frame_id = ODOM_FRAME_ID;

			odom.pose.pose.position.x = x;
			odom.pose.pose.position.y = y;
			odom.pose.pose.position.z = 0.0;
			odom.pose.pose.orientation = odomQuat;

			odom.child_frame_id = CHILD_FRAME_ID;
			odom.twist.twist.linear.x = v[0];
			odom.twist.twist.linear.y = v[1];
			odom.twist.twist.angular.z = omega = w.z();

			w_sum[0].setZero();
			sensingSteps[0] = 0;

			updateLocalization();

			ros::spinOnce();
			odom_pub_.publish(odom);

			std_msgs::Float32MultiArray array;
			array.data.resize(6);
			array.data[0] = 0.0;
			array.data[1] = 0.0;
			array.data[2] = th;
			array.data[3] = 0.0;
			array.data[4] = 0.0;
			array.data[5] = (th * 180) / M_PI;
			theta_pub_.publish(array);

			lastTime[0] += update_period[0];

			ROS_INFO("-------------------- Start output of odometry --------------------");
			ROS_INFO("position: ( %lf, %lf, %lf )", x, y, 0.0);
			ROS_INFO("orientation: ( %lf, %lf, %lf )", 0.0, 0.0, odomQuat.z);
			ROS_INFO("linear: ( %lf, %lf, %lf )", v[0], v[1], 0.0);
			ROS_INFO("angular: ( %lf, %lf, %lf )", 0.0, 0.0, omega);
			ROS_INFO("-------------------- End output of odometry --------------------");
		}
	}

	void updateImu()
	{
		currentTime[1] = io->currentTime();
		double updateSince = currentTime[1] - lastTime[1];

		dv_sum[1] += accel->dv();
		w_sum[1] += gyro->w();
		sensingSteps[1]++;

		if(updateSince > update_period[1]){
			auto dv = dv_sum[1] / sensingSteps[1];
			auto w = w_sum[1] / sensingSteps[1];
			sensor_msgs::Imu imu;

			imuTh += w.z() * update_period[1];

			geometry_msgs::Quaternion imuQuat = tf::createQuaternionMsgFromYaw(imuTh);
			imu.header.stamp.fromSec(currentTime[1]);
			imu.header.frame_id = IMU_FRAME_ID;
			imu.linear_acceleration.x = dv.x();
			imu.linear_acceleration.y = dv.y();
			imu.linear_acceleration.z = dv.z();
			imu.angular_velocity.x = w.x();
			imu.angular_velocity.y = w.y();
			imu.angular_velocity.z = w.z();
			imu.orientation = imuQuat;

			dv_sum[1].setZero();
			w_sum[1].setZero();
			sensingSteps[1] = 0;

			ros::spinOnce();
			imu_pub_.publish(imu);

			lastTime[1] += update_period[1];

			ROS_INFO("-------------------- Start output of imu --------------------");
			ROS_INFO("Accel Sensor  dvx: %lf, dvy: %lf, dvz: %lf", dv.x(), dv.y(), dv.z());
			ROS_INFO("Gyro Sensor  wx: %lf, wy: %lf, wz: %lf", w.x(), w.y(), w.z());
			ROS_INFO("-------------------- End output of imu --------------------");
		}
	}

	virtual bool control() override
	{
		sensor_msgs::Joy joystick;
		{
			std::lock_guard<std::mutex> lock(mutex);
			joystick = latestJoystickState;
			joystick.axes.resize(10, 0.0f);
			joystick.buttons.resize(10, 0);
		}

		double pos[2];		// Amount of change in joystick.
		for(int i = 0; i < 2; ++i){
			// Get joystick value.
			pos[i] = joystick.axes[
					i == 0 ? cnoid::Joystick::DIRECTIONAL_PAD_H_AXIS : cnoid::Joystick::DIRECTIONAL_PAD_V_AXIS];
//					i == 0 ? cnoid::Joystick::L_STICK_H_AXIS : cnoid::Joystick::L_STICK_V_AXIS];
			if(fabs(pos[i]) < 0.2){
				// When absolute value is less than 0.2.
				pos[i] = 0.0;
			}
		}

		if(actuationMode == cnoid::Link::JointVelocity){
			// When actuation mode is velocity.
			static const double K = 6.0;
			wheels[0]->dq_target() = -K * (pos[1] - 0.5 * pos[0]);
			wheels[1]->dq_target() = -K * (pos[1] + 0.5 * pos[0]);

		} else if(actuationMode == cnoid::Link::JointTorque){
			// When actuation mode is torque.
			static const double K1 = 0.3;
			static const double K2 = 0.3;
			double dq_target[2];		// Target angular velocity.
			dq_target[0] = -K1 * pos[1] + K2 * pos[0];		// pos[1]: -1(UP), 1(DOWN)
			dq_target[1] = -K1 * pos[1] - K2 * pos[0];		// pos[0]: -1(LEFT), 1(RIGHT)

			for(int i = 0; i < 2; ++i){
				double dq_current = (wheels[i]->q() - qprev[i]) / dt;	// Current angular velocity.
				wheels[i]->u() = (dq_target[i] - dq_current);	// Torque value given to tire.
				qprev[i] = wheels[i]->q();	// Update previous joint angular.
			}
		}

		updateImu();
		updateOdom();

		return true;
	}

	virtual void stop() override
	{
		x = y = th = 0.0;
		for(int i = 0; i < sizeof(v)/sizeof(v[0]); ++i){
			v[i] = 0.0;
		}

		joy_sub_.shutdown();
		cmd_vel_sub_.shutdown();
		odom_pub_.shutdown();
		imu_pub_.shutdown();
		theta_pub_.shutdown();
		pcl_sub_.shutdown();
		pcl2_pub_.shutdown();
	}
};

CNOID_IMPLEMENT_SIMPLE_CONTROLLER_FACTORY(TurtleBot3RosController)
