/*
 * TurtleBot3RosController.cpp
 *
 *  Created on: 2020/03/25
 *      Author: Tsuyoshi Anazawa
 *
 *  Outline: The controller for operating TurtleBot3 model with PS4 controller.
 */

#include <cnoid/SimpleController>
#include <cnoid/RangeCamera>
#include <cnoid/RangeSensor>
#include <cnoid/AccelerationSensor>
#include <cnoid/RateGyroSensor>
#include <cnoid/Joystick>
#include <fmt/format.h>
#include <ros/node_handle.h>
#include <sensor_msgs/Joy.h>
#include <sensor_msgs/Imu.h>
#include <mutex>
#include <iostream>
#include <ros/ros.h>
#include <nav_msgs/Odometry.h>
#include <tf/transform_broadcaster.h>
#include <geometry_msgs/Twist.h>
#include <geometry_msgs/TransformStamped.h>
// For angle confirmation(Debug).
#include <std_msgs/Float32MultiArray.h>
#include <sensor_msgs/PointCloud.h>
#include <sensor_msgs/PointCloud2.h>
#include <sensor_msgs/point_cloud_conversion.h>

using namespace std;
using namespace cnoid;
using fmt::format;

class TurtleBot3RosController : public SimpleController
{
	ros::NodeHandle node;
	// Joystick subscriber
	ros::Subscriber joystickSubscriber;
	sensor_msgs::Joy latestJoystickState;
	std::mutex joystickMutex;
	// Odometry publisher
	ros::Publisher odomPublisher;
	tf::TransformBroadcaster tfBroadcaster;
	// 0: odom, 1: imu, 2: cmd_vel
	double currentTime[3], lastTime[3], update_period[3];
	double update_rate[3] = { 3.0, 20.0, 10.0 };

	// Imu publisher
	ros::Publisher imuPublisher;
	AccelerationSensorPtr accelSensor;
	RateGyroSensorPtr gyroSensor;
	string deviceName[2] = { "AccelSensor", "GyroSensor" };
	ScopedConnection accelSensorConnection;
	ScopedConnection gyroSensorConnection;

	// cmd_vel publisher
	ros::Publisher cmdvelPublisher;
	// cmd_vel subscriber
	ros::Subscriber cmdvelSubscriber;
	// For angle confirmation(Debug)
	ros::Publisher thetaPublisher;
	ros::Publisher pointcloud2pub_;
	ros::Subscriber pointcloudsub_;
	// Constant to store the number of wheels.
	static const int WHEEL_NUM = 2;
	// The array containing the wheel names defined in the body file.
	const string wheelNames[WHEEL_NUM] = { "Left_wheel", "Right_wheel" };
	// The variable that stores the actuation mode of the link.
	int actuationMode;
	// The array to store wheels.
	Link* wheels[2];
	double qprev[2];
	double dt;
	double qref[2];
	SimpleControllerIO* io;
	// Current coordinate position.
	double x, y, th;
	// The array used in accelerometers and angular velocity sensors.
	// Current velocity[m/s] and angular velocity[rad/s].
	Vector3 v, vth;
	// Last velocity[m/s].
	Vector3 prevV, prevVth;
	// Last angular velocity[m/s*s].
	Vector3 prevDv;
	Vector3 calcV_odom, calcV_cmdvel, calcVth_odom, calcVth_cmdvel;
	Vector3 cmdvel, cmdvelTh;
	// The variable that stores Low Pass Filter and High Pass Filter.
	double lpf, hpf;
	// Low Pass Filter coefficient.
	double alpha = 0.9;
	double timeSpan = 0.05;
//	const string ODOM_FRAME_ID = "waffle_pi/odom";
//	const string CHILD_FRAME_ID = "waffle_pi/base_footprint";
//	const string IMU_FRAME_ID = "waffle_pi/base_footprint";
	const string ODOM_FRAME_ID = "odom";
	const string CHILD_FRAME_ID = "base_footprint";
	const string IMU_FRAME_ID = "base_footprint";
	double d = 0.1435;

public:

	virtual bool configure(SimpleControllerConfig* config) override
	{
		return true;
	}

	virtual bool initialize(SimpleControllerIO* io) override
	{
		this->io = io;
		ostream& os = io->os();
		// Body variable.
		Body* body = io->body();

		// Set joint's actuation mode.
		actuationMode = Link::JointTorque;
		// Get controller option.
		string option = io->optionString();

		if(!option.empty()){
			// When controller option is empty.
			if(option == "velocity" || option == "position"){
				// When controller option is velocity or position.
				actuationMode = Link::JointVelocity;
			} else if(option == "torque"){
				// When controller option is torque.
				actuationMode = Link::JointTorque;
			} else {
				// In other cases.
				os << format("Warning: Unknown option {}.", option) << endl;
			}
		}

		for(int i = 0; i < WHEEL_NUM; ++i){
			// Store wheel links in array.
			wheels[i] = body->link(wheelNames[i]);
			if(!wheels[i]){
				// When there is no wheel links.
				os << format("{0} of {1} is not found.", wheelNames[i], body->name()) << endl;
				return false;
			}

			// Set the wheel array's actuation mode.
			wheels[i]->setActuationMode(actuationMode);
			// Enable output to wheel links.
			io->enableOutput(wheels[i]);
			// Store current joint angles in array
			qprev[i] = wheels[i]->q();
		}

		accelSensor = body->findDevice<AccelerationSensor>(deviceName[0]);
		if(!accelSensor){
			os << format("{0} of {1} is not found.", deviceName[0], body->name()) << endl;
			//return false;
		} else {
			io->enableInput(accelSensor);
			accelSensorConnection.disconnect();
			accelSensorConnection = accelSensor->sigStateChanged().connect(
					[&](){ updateImu(); });
		}
		
		gyroSensor = body->findDevice<RateGyroSensor>(deviceName[1]);
		if(!gyroSensor){
			os << format("{0} of {1} is not found.", deviceName[1], body->name()) << endl;
			//return false;
		} else {
			io->enableInput(gyroSensor);
			gyroSensorConnection.disconnect();
			gyroSensorConnection = gyroSensor->sigStateChanged().connect(
					[&](){ updateImu(); });
		}

		// joystick subscriber
		joystickSubscriber = node.subscribe("joy", 1, &TurtleBot3RosController::joystickCallback, this);
		// cmd_vel subscriber
		cmdvelSubscriber = node.subscribe("waffle_pi/cmd_vel", 1, &TurtleBot3RosController::cmdvelCallback, this);
		// Create the publisher of odometry(topic: odom, buffer_size: 10).
//		odomPublisher = node.advertise<nav_msgs::Odometry>("waffle_pi/odom", 10);
		odomPublisher = node.advertise<nav_msgs::Odometry>("odom", 10);
		// Create the publisher of imu(topic: imu, buffer_size: 10).
//		imuPublisher = node.advertise<sensor_msgs::Imu>("waffle_pi/imu", 10);
		imuPublisher = node.advertise<sensor_msgs::Imu>("imu", 10);
		// Create the publisher of cmd_vel(topic: cmd_vel, buffer_size: 10).
//		cmdvelPublisher = node.advertise<geometry_msgs::Twist>("waffle_pi/cmd_vel", 10);
		cmdvelPublisher = node.advertise<geometry_msgs::Twist>("cmd_vel", 10);
		// For angle confirmation(Debug).
		thetaPublisher = node.advertise<std_msgs::Float32MultiArray>("waffle_pi/theta", 10);
		pointcloudsub_ = node.subscribe("waffle_pi/velodyne/point_cloud", 1, &TurtleBot3RosController::pointcloudCallback, this);
//		pointcloud2pub_ = node.advertise<sensor_msgs::PointCloud2>("velodyne_points", 10);
		pointcloud2pub_ = node.advertise<sensor_msgs::PointCloud2>("multi_scan_points", 10);

		// Suppose the robot starts from the origin of the odom coordinate system.
		x = 0.0;
		y = 0.0;
		th = 0.0;

		prevV[0] = prevV[1] = prevVth[2] = 0.0;
		// The acceleration and the angular velocity for odom and cmd_vel calculations.
		for(int i = 0; i < sizeof(calcV_odom) / sizeof(calcV_odom[0]); ++i){
			calcV_odom[i] = 0.0;
			calcV_cmdvel[i] = 0.0;
			calcVth_odom[i] = 0.0;
			calcVth_cmdvel[i] = 0.0;
		}
		lpf = hpf = 0.0;

		currentTime[3] = lastTime[3] = io->currentTime();
		for(int i = 0; i < sizeof(currentTime) / sizeof(currentTime[0]); ++i){
			update_period[i] = 1.0 / update_rate[i];
		}

		dt = io->timeStep();

		return true;
	}

	void joystickCallback(const sensor_msgs::Joy& msg)
	{
		std::lock_guard<std::mutex> lock(joystickMutex);
		latestJoystickState = msg;
	}

	void cmdvelCallback(const geometry_msgs::Twist& msg)
	{
		cmdvel[0] = msg.linear.x;
		cmdvel[1] = msg.linear.y;
		cmdvel[2] = msg.linear.z;
		cmdvelTh[0] = msg.angular.x;
		cmdvelTh[1] = msg.angular.y;
		cmdvelTh[2] = msg.angular.z;
		ROS_INFO("------------------ Subscribe of cmd_vel ------------------");
		ROS_INFO("linear: ( %lf, %lf, %lf )", msg.linear.x, msg.linear.y, msg.linear.z);
		ROS_INFO("angular: ( %lf, %lf, %lf )", msg.angular.x, msg.angular.y, msg.angular.z);
		ROS_INFO("------------------ Subscribe of cmd_vel ------------------");
	}

	void pointcloudCallback(const sensor_msgs::PointCloud& msg)
	{
		sensor_msgs::PointCloud2 pc2_out;
		sensor_msgs::PointCloud pc_in;
//		pc_in = msg;
		pc_in.header = msg.header;
		pc_in.channels = msg.channels;
//		ROS_INFO("%d", msg.points.size());
		for(int i = 0; i < msg.points.size(); i++){
//			ROS_INFO("%lf  %lf  %lf", msg.points[i].x, msg.points[i].y, msg.points[i].z);
			geometry_msgs::Point32 point;
			point.x = -msg.points[i].z;
			point.y = -msg.points[i].x;
			point.z = msg.points[i].y;
			pc_in.points.push_back(point);
		}
		sensor_msgs::convertPointCloudToPointCloud2(pc_in, pc2_out);
		pointcloud2pub_.publish(pc2_out);
//		pc2_out.data[0];
	}

	void updateOdom()
	{
		// odom座標系でx, y, θ方向への速度設定
//		vx = (wheels[0]->dq() + wheels[1]->dq()) / 2;
//		vy = (wheels[0]->dq() - wheels[1]->dq()) / 2;
//		vth = (wheels[0]->dq() - wheels[1]->dq()) / 2 * d;

		currentTime[0] = io->currentTime();
		double updateSince = currentTime[0] - lastTime[0];

		if(updateSince > update_period[0]){
			currentTime[0] = io->currentTime();
			// imu(20Hz)とodom(3.0Hz)と周期が異なるので再計算
//			v[0] = (v[0] * update_period[0]) / update_period[1];
//			v[1] = (v[1] * update_period[0]) / update_period[1];
//			calcSpeed(0, calcDv_odom);
//			calcSpeed(1, calcDv_odom);
//			vth[2] = calcVth_odom[2];
			double deltaT = (currentTime[0] - lastTime[0]);
			double deltaX = (v[0] * cos(th) - v[1] * sin(th)) * deltaT;
			double deltaY = (v[0] * sin(th) + v[1] * cos(th)) * deltaT;
			double deltaTh = (vth[2] + prevVth[2]) * deltaT / 2.0;
			// 修正箇所
//			calcV_odom[0] = calcV_odom[0] / (update_period[0] / update_period[1]);
//			calcV_odom[1] = calcV_odom[1] / (update_period[0] / update_period[1]);
//			calcVth_odom[2] = calcVth_odom[2] / (update_period[0] / update_period[1]);
//			double deltaX = (calcV_odom[0] * cos(th) - calcV_odom[1] * sin(th)) * deltaT;
//			double deltaY = (calcV_odom[0] * sin(th) + calcV_odom[1] * cos(th)) * deltaT;
//			double deltaTh = (calcVth_odom[2] + prevVth[2]) * deltaT / 2.0;

			x += deltaX;
			y += deltaY;
			th += deltaTh;
			// Convert yaw values to quaternions.
			geometry_msgs::Quaternion odomQuat = tf::createQuaternionMsgFromYaw(th);
			// Create a TransformedStamped message to send to tf.
			geometry_msgs::TransformStamped odomTrans;
			odomTrans.header.stamp.fromSec(currentTime[0]);
			odomTrans.header.frame_id = ODOM_FRAME_ID;
			odomTrans.child_frame_id = CHILD_FRAME_ID;

			// Substitute coordinate transformation data from odometry data and send using Transform Broadcaster.
			odomTrans.transform.translation.x = x;
			odomTrans.transform.translation.y = y;
			odomTrans.transform.translation.z = 0.0;
			odomTrans.transform.rotation = odomQuat;
			tfBroadcaster.sendTransform(odomTrans);

			// The navigation stack publishes a nav_msgs/Odometry message for velocity information.
			nav_msgs::Odometry odom;
			odom.header.stamp.fromSec(currentTime[0]);
			odom.header.frame_id = ODOM_FRAME_ID;

			odom.pose.pose.position.x = x;
			odom.pose.pose.position.y = y;
			odom.pose.pose.position.z = 0.0;
			odom.pose.pose.orientation = odomQuat;

			odom.child_frame_id = CHILD_FRAME_ID;
			odom.twist.twist.linear.x = v[0];
			odom.twist.twist.linear.y = v[1];
			odom.twist.twist.angular.z = vth[2];

			ros::spinOnce();
			odomPublisher.publish(odom);

			std_msgs::Float32MultiArray array;
			array.data.resize(6);
			array.data[0] = 0.0;
			array.data[1] = 0.0;
			array.data[2] = th;
			array.data[3] = 0.0;
			array.data[4] = 0.0;
			array.data[5] = (th * 180) / M_PI;
			thetaPublisher.publish(array);

			lastTime[0] += update_period[0];
			prevVth[2] = vth[2];
			calcV_odom[0] = calcV_odom[1] = calcVth_odom[2] = 0.0;
//			calcDv_odom[0] = calcDv_odom[1] = calcVth_odom[2] = 0.0;
//			v[0] = v[1] = 0.0;

			ROS_INFO("-------------------- Start output of odometry --------------------");
			ROS_INFO("position: ( %lf, %lf, %lf )", x, y, 0.0);
			ROS_INFO("orientation: ( %lf, %lf, %lf )", 0.0, 0.0, odomQuat.z);
			ROS_INFO("linear: ( %lf, %lf, %lf )", v[0], v[1], 0.0);
			ROS_INFO("angular: ( %lf, %lf, %lf )", 0.0, 0.0, vth[2]);
			ROS_INFO("-------------------- End output of odometry --------------------");
		}
	}

	void updateImu()
	{
		currentTime[1] = io->currentTime();
		double updateSince = currentTime[1] - lastTime[1];

		if(updateSince > update_period[1]){
			currentTime[1] = io->currentTime();
			auto dv = accelSensor->dv();
			auto w = gyroSensor->w();
			sensor_msgs::Imu imu;

			// x velocity.s
			calcSpeed(0, dv);
			// y velocity.
			calcSpeed(1, dv);
			vth[2] = w.z();
//			calcVth_odom[2] += w.z();
//			calcVth_cmdvel[2] += w.z();

			imu.header.stamp.fromSec(currentTime[1]);
			imu.header.frame_id = IMU_FRAME_ID;
			imu.linear_acceleration.x = dv.x();
			imu.linear_acceleration.y = dv.y();
			imu.linear_acceleration.z = dv.z();
			imu.angular_velocity.x = w.x();
			imu.angular_velocity.y = w.y();
			imu.angular_velocity.z = w.z();

			ros::spinOnce();
			imuPublisher.publish(imu);

			lastTime[1] += update_period[1];

			ROS_INFO("-------------------- Start output of imu --------------------");
			ROS_INFO("Accel Sensor  dvx: %lf, dvy: %lf, dvz: %lf", dv.x(), dv.y(), dv.z());
			ROS_INFO("Gyro Sensor  wx: %lf, wy: %lf, wz: %lf", w.x(), w.y(), w.z());
			ROS_INFO("-------------------- End output of imu --------------------");
		}
	}

	void calcSpeed(int id, Vector3 dv)
	{
		if(wheels[0]->dq_target() != 0.0 && wheels[1]->dq_target() != 0.0 ){
			// ローパスフィルタ(現在値 = 係数 * 前回値 + (1 - 係数) * センサ値)
			lpf = alpha * lpf + (-dv[id]) * (1 - alpha);
			// ハイパスフィルタ(現在の値 = センサ値 - ローパスフィルタの値)
			hpf = (-dv[id]) - lpf;

			// 速度計算(加速度を台形積分)
			// 修正箇所
			v[id] = ((hpf + prevDv[id]) * update_period[1]) / 2.0 + v[id];
//			calcV_odom[id] += (hpf + prevDv[id] * update_period[1]) / 2.0;
//			calcV_cmdvel[id] += (hpf + prevDv[id] * update_period[1]) / 2.0;
//			v[id] = ((hpf + prevDv[id]) / update_period[2]) / 2.0;
//			v[id] = ((dv[id] + prevDv[id]) * timeSpan) / 2.0 + prevV[id];
//			v[id] = -(dv[id] + prevDv[id]) / 2.0;
 			prevDv[id] = hpf;
//			prevDv[id] = dv[id];
//			prevV[id] = v[id];
		} else {
			v[id] = dv[id] = prevDv[id] = 0.0;
		}
	}

	void updateCmdVel()
	{
		geometry_msgs::Twist cmdVel;

		currentTime[2] = io->currentTime();
		double updateSince = currentTime[2] - lastTime[2];
		if(updateSince > update_period[2]){
			// 修正箇所
//			calcV_cmdvel[0] = calcV_cmdvel[0] / (update_period[2] / update_period[1]);
//			calcV_cmdvel[1] = calcV_cmdvel[1] / (update_period[2] / update_period[1]);
//			calcVth_cmdvel[2] = calcVth_cmdvel[2] / (update_period[2] / update_period[1]);

			cmdVel.linear.x = v[0];
			cmdVel.linear.y = v[1];
//			cmdVel.linear.x = calcV_cmdvel[0];
//			cmdVel.linear.y = calcV_cmdvel[1];
			cmdVel.linear.z = 0.0;
			cmdVel.angular.x = 0.0;
			cmdVel.angular.y = 0.0;
			cmdVel.angular.z = vth[2];

			ros::spinOnce();
			cmdvelPublisher.publish(cmdVel);

			lastTime[2] += update_period[2];
			calcV_cmdvel[0] = calcV_cmdvel[1] = calcVth_cmdvel[2] = 0.0;

//			ROS_INFO("-------------------- Start output of cmd_vel --------------------");
//			ROS_INFO("linear: ( %lf, %lf, %lf )", cmdVel.linear.x, cmdVel.linear.y, cmdVel.linear.z);
//			ROS_INFO("angular: ( %lf, %lf, %lf )", cmdVel.angular.x, cmdVel.angular.y, cmdVel.angular.z);
//			ROS_INFO("-------------------- End output of cmd_vel --------------------");
		}
	}

	virtual bool control() override
	{
		updateOdom();
		sensor_msgs::Joy joystick;
		{
			std::lock_guard<std::mutex> lock(mutex);
			joystick = latestJoystickState;
			joystick.axes.resize(10, 0.0f);
			joystick.buttons.resize(10, 0);
		}

		double pos[2];		// Amount of change in joystick.
		for(int i = 0; i < 2; ++i){
			// Get joystick value.
			pos[i] = joystick.axes[
					i == 0 ? Joystick::DIRECTIONAL_PAD_H_AXIS : Joystick::DIRECTIONAL_PAD_V_AXIS];
			if(fabs(pos[i]) < 0.2){
				// When absolute value is less than 0.2.
				pos[i] = 0.0;
			}
		}

		if(actuationMode == Link::JointVelocity){
			// When actuation mode is velocity.
//			static const double K = 8.8;
			static const double K = 6.0;
//			if(pos[1] != 0.0 && pos[0] != 0.0){
				wheels[0]->dq_target() = -K * (pos[1] - 0.5 * pos[0]);
				wheels[1]->dq_target() = -K * (pos[1] + 0.5 * pos[0]);
//			} else {
//				if(cmdvel[0] >= 0.01 || cmdvelTh[2] >= 0.001){
//					wheels[0]->dq_target() = 5.0 * (cmdvel[0] / 0.26 - (cmdvelTh[2] / 0.26) * d);
//					wheels[1]->dq_target() = 5.0 * (cmdvel[0] / 0.26 + (cmdvelTh[2] / 0.26) * d);
//				} else {
//					wheels[0]->dq_target() = 0.0;
//					wheels[1]->dq_target() = 0.0;
//				}
//			}
//			ROS_INFO("control function:\tv: %lf", (wheels[0]->dq_target() + wheels[1]->dq_target()) / 2.0);

		} else if(actuationMode == Link::JointTorque){
			// When actuation mode is torque.
			static const double K1 = 0.14;
			static const double K2 = 0.25;
			double dq_target[2];		// Target angular velocity.
			dq_target[0] = -K1 * pos[1] + K2 * pos[0];		// pos[1]: -1(UP), 1(DOWN)
			dq_target[1] = -K1 * pos[1] - K2 * pos[0];		// pos[0]: -1(LEFT), 1(RIGHT)

			for(int i = 0; i < 2; ++i){
				double dq_current = (wheels[i]->q() - qprev[i]) / dt;	// Current angular velocity.
				wheels[i]->u() = (dq_target[i] - dq_current);	// Torque value given to tire.
				qprev[i] = wheels[i]->q();	// Update previous joint angular.
			}
		}
		updateCmdVel();

		return true;
	}

	virtual void stop() override
	{
		x = y = th = 0.0;
		for(int i = 0; i < sizeof(v)/sizeof(v[0]); ++i){
			v[i] = vth[i] = prevV[i] = prevDv[0] = 0.0;
		}

		joystickSubscriber.shutdown();
		cmdvelSubscriber.shutdown();
		odomPublisher.shutdown();
		imuPublisher.shutdown();
		cmdvelPublisher.shutdown();
	}
};

CNOID_IMPLEMENT_SIMPLE_CONTROLLER_FACTORY(TurtleBot3RosController)
