/*
 * TurtleBot3RosNaviController.cpp
 *
 *  Created on: 2021/04/01
 *      Author: Tsuyoshi Anazawa
 *
 *  Outline: This program that controls the TurtleBot3 model with ROS navigation.
 */

#include <cnoid/SimpleController>
#include <cnoid/RangeCamera>
#include <cnoid/RangeSensor>
#include <cnoid/AccelerationSensor>
#include <cnoid/RateGyroSensor>
#include <fmt/format.h>
#include <ros/node_handle.h>
#include <sensor_msgs/Imu.h>
#include <mutex>
#include <iostream>
#include <ros/ros.h>
#include <nav_msgs/Odometry.h>
#include <tf/transform_broadcaster.h>
#include <geometry_msgs/Twist.h>
// For angle confirmation(Debug).
#include <std_msgs/Float32MultiArray.h>
// simulation time
#include <std_msgs/String.h>
#include <cnoid/EigenUtil>
#include <math.h>
// Output csv file.
#include <fstream>

using namespace std;
using namespace cnoid;
using fmt::format;

class TurtleBot3RosNaviController : public SimpleController
{
	ros::NodeHandle nh_;
	// Publisher for Odometry
	ros::Publisher odom_pub_;
	tf::TransformBroadcaster tfBroadcaster_;
	// 0: odom, 1: imu, 2: cmd_vel
	double currentTime[3], lastTime[3], update_period[3];
	double update_rate[3] = { 30.0, 200.0, 10.0 };

	// Publisher for imu
	ros::Publisher imu_pub_;
	AccelerationSensorPtr accelSensor;
	RateGyroSensorPtr gyroSensor;
	string deviceName[2] = { "AccelSensor", "GyroSensor" };
	ScopedConnection accelSensorConnection;
	ScopedConnection gyroSensorConnection;

	// Subscriber for cmd_vel
	ros::Subscriber cmd_vel_sub_;
	// Publisher for angle[deg](Debug)
	ros::Publisher theta_pub_;
	// Number of wheels
	static const int WHEEL_NUM = 2;
	// Wheel names
	const string wheelNames[WHEEL_NUM] = { "Left_wheel", "Right_wheel" };
	// Actuation mode
	int actuationMode;
	// Wheel links
	Link* wheels[2];
	double qprev[2];
	double dt;
	double qref[2];
	SimpleControllerIO* io;
	// current coordinate position
	double x, y, th;
	double omega;
//	double debug_th;
	// Current velocities[m/s] and angular velocities[rad/s]
	Vector3 v, vth;
	// Last velocities[m/s] and angular velocities[rad/s]
	Vector3 prevVth, debug_vth;
	double imuTh;
	Vector3 cmd_vel, cmd_vel_th;
	double timeSpan = 0.05;
	// Odometry and imu frame id
	const string ODOM_FRAME_ID = "tb3_01/odom";
	const string CHILD_FRAME_ID = "tb3_01/base_footprint";
	const string IMU_FRAME_ID = "tb3_01/base_footprint";
	// Limit velocities and accelerations
	const double MAX_VEL = 0.26;
	const double MIN_VEL = -0.26;
	const double MAX_ACCEL = 1.82;
	const double MIN_ACCEL = -1.82;
	// half of the tread
	double d = 0.1435;
	// Wheel radius
	double r = 0.033;
	Vector2 prevWheelTh;
	Vector3 dv_sum[2], w_sum[2];
	int sensingSteps[2];
	double wheel_dq[2];
	// 2022/5/19 Add for debug.
	double x_deb, y_deb, th_deb;
	Vector3 v_deb;

	// rootLink
	BodyPtr ioBody;
	double count = 0;
	double ptheta;
	// Simulation time.
	ros::Publisher simtime_pub_;
	// Output csv file.
	ofstream ofs;
	double current_time, last_time;

public:
	double clamp(const double x, const double min, const double max)
	{
	  return std::min(std::max(min, x), max);
	}

	virtual bool configure(SimpleControllerConfig* config) override
	{
		return true;
	}

	virtual bool initialize(SimpleControllerIO* io) override
	{
		this->io = io;
		ostream& os = io->os();
		// Body variable.
		Body* body = io->body();
		// Add to actual value.
		ioBody = body;

		// Set joint's actuation mode.
		actuationMode = Link::JointTorque;
		// Get controller option.
		string option = io->optionString();

		if(!option.empty()){
			// When controller option is empty.
			if(option == "velocity" || option == "position"){
				// When controller option is velocity or position.
				actuationMode = Link::JointVelocity;
			} else if(option == "torque"){
				// When controller option is torque.
				actuationMode = Link::JointTorque;
			} else {
				// In other cases.
				os << format("Warning: Unknown option {}.", option) << endl;
			}
		}

		// Add to actual value.
		io->enableInput(ioBody->rootLink(), LINK_POSITION);

		for(int i = 0; i < WHEEL_NUM; ++i){
			// Store wheel links in array.
			wheels[i] = body->link(wheelNames[i]);
			if(!wheels[i]){
				// When there is no wheel links.
				os << format("{0} of {1} is not found.", wheelNames[i], body->name()) << endl;
				return false;
			}

			// Set the wheel array's actuation mode.
			wheels[i]->setActuationMode(actuationMode);
			// Enable output to wheel links.
			io->enableIO(wheels[i]);
			// Store current joint angles in array
			qref[i] = qprev[i] = wheels[i]->q();
		}

		accelSensor = body->findDevice<AccelerationSensor>(deviceName[0]);
		if(!accelSensor){
			os << format("{0} of {1} is not found.", deviceName[0], body->name()) << endl;
			//return false;
		} else {
			io->enableInput(accelSensor);
//			accelSensorConnection.disconnect();
//			accelSensorConnection = accelSensor->sigStateChanged().connect(
//					[&](){ updateImu(); });
		}
		
		gyroSensor = body->findDevice<RateGyroSensor>(deviceName[1]);
		if(!gyroSensor){
			os << format("{0} of {1} is not found.", deviceName[1], body->name()) << endl;
			//return false;
		} else {
			io->enableInput(gyroSensor);
//			gyroSensorConnection.disconnect();
//			gyroSensorConnection = gyroSensor->sigStateChanged().connect(
//					[&](){ updateImu(); });
		}

		// Create subscriber for cmd_vel
		cmd_vel_sub_ = nh_.subscribe("cmd_vel", 1, &TurtleBot3RosNaviController::cmdvelCallback, this);
		// Create publisher for odometry(topic: odom, buffer_size: 10)
		odom_pub_ = nh_.advertise<nav_msgs::Odometry>("odom", 10);
		// Create publisherfor imu(topic: imu, buffer_size: 10)
		imu_pub_ = nh_.advertise<sensor_msgs::Imu>("imu", 10);
		// Create publisher for angle(Debug)
		theta_pub_ = nh_.advertise<std_msgs::Float32MultiArray>("theta", 10);
		// Create publisher for simulation time.
		simtime_pub_ = nh_.advertise<std_msgs::String>("simtime", 10);
		// Origin of odom coordinate system assuming the initial position of the robot
		double init_pose_x, init_pose_y, init_pose_theta;
		x = nh_.param("tb3_01_x_pos", init_pose_x, 0.0);
		y = nh_.param("tb3_01_y_pos", init_pose_y, 0.0);
		th = imuTh = nh_.param("tb3_01_yaw", init_pose_theta, 0.0);
		// 2022/5/19 Add for debug.
		x_deb = nh_.param("tb3_01_x_pos", init_pose_x, 0.0);
		y_deb = nh_.param("tb3_01_y_pos", init_pose_y, 0.0);
		th_deb = imuTh = nh_.param("tb3_01_yaw", init_pose_theta, 0.0);
		ROS_INFO("x: %lf. y: %lf, yaw: %lf",x, y, th);
		// Add to actual value.
		ptheta = init_pose_theta;

		prevVth[2] = 0.0;
		debug_vth[2] = 0.0;

		for(int i = 0; i < sizeof(wheels) / sizeof(wheels[0]); ++i){
			prevWheelTh[i] = wheels[i]->q();
		}

		currentTime[3] = lastTime[3] = io->currentTime();
		for(int i = 0; i < sizeof(currentTime) / sizeof(currentTime[0]); ++i){
			update_period[i] = 1.0 / update_rate[i];
		}

		string file_path = "/home/fsk9001/.ros/cnoid_coord_navi.csv";
		ifstream ifs(file_path);
		string hdr;
		bool hdr_flg = false;
		if(ifs.fail()){
			hdr = "sim_time,x,y,z";
			hdr_flg = true;
		}
		ofs.open(file_path, ios::app);
		if(hdr_flg){
			ofs << hdr << endl;
		}

		current_time = last_time = io->currentTime();
		dt = io->timeStep();
		sensingSteps[0] = sensingSteps[1] = 0;

		return true;
	}

	void cmdvelCallback(const geometry_msgs::Twist& msg)
	{
		cmd_vel[0] = msg.linear.x;
		cmd_vel[1] = msg.linear.y;
		cmd_vel[2] = msg.linear.z;
		cmd_vel_th[0] = msg.angular.x;
		cmd_vel_th[1] = msg.angular.y;
		cmd_vel_th[2] = msg.angular.z;
	}

	void updateCoord()
	{
		Vector2 wheel_th, wheel_omg, wheel_v;

		for(int i = 0; i < sizeof(wheels) / sizeof(wheels[0]); ++i){
			double latest_wheel_angle = wheels[i]->q();
			wheel_th[i] = (latest_wheel_angle - prevWheelTh[i]);
//			wheel_omg[i] = wheel_th[i] / dt;
			/* 2022/05/18 座標位置計算方法の変更時の検証 */
//			wheel_omg[i] = wheel_th[i] / update_period[0];
			wheel_omg[i] = wheel_dq[i];

//			std::cout << "Before wheel angular velocity: " << wheel_th[i] / update_period[0] << std::endl;
//			std::cout << "After wheel angular velocity: " << wheel_dq[i] << std::endl;
			wheel_v[i] = r * wheel_omg[i];
			string w_nm;
			if(i == 0){
				w_nm = "left";
			} else {
				w_nm = "right";
			}
			prevWheelTh[i] = latest_wheel_angle;
		}
		v[0] = clamp((wheel_v[0] + wheel_v[1]) / 2.0, MIN_VEL, MAX_VEL);
		v[1] = 0.0;

//		omega = clamp((wheel_v[1] - wheel_v[0]) / (2.0 * d), MIN_ACCEL, MAX_ACCEL);
//		omega = gyroSensor->w().z();
//		double deltaTh = ((omega + prevVth[2]) / 2.0) * dt;
//		double deltaX = (v[0] * cos(th) - v[1] * sin(th)) * dt;
//		double deltaY = (v[0] * sin(th) + v[1] * cos(th)) * dt;
		double deltaX = (v[0] * cos(th) - v[1] * sin(th)) * update_period[0];
		double deltaY = (v[0] * sin(th) + v[1] * cos(th)) * update_period[0];
		double deltaTh = ((omega + prevVth[2]) / 2.0) * update_period[0];
//		double deltaTh = ((omega + prevVth[2]) / 2.0) * dt;

		x += deltaX;
		y += deltaY;
		th += deltaTh;

		// 2022/5/19 Add for debug.
		v_deb[0] = clamp((r * wheel_th[0] / update_period[0] + r * wheel_th[1] / update_period[0]) / 2.0, MIN_VEL, MAX_VEL);
		v_deb[1] = 0.0;
		double delta_x_deb = (v_deb[0] * cos(th) - v_deb[1] * sin(th)) * update_period[0];
		double delta_y_deb = (v_deb[0] * sin(th) + v_deb[1] * cos(th)) * update_period[0];
		double delta_th_deb = ((omega + prevVth[2]) / 2.0) * update_period[0];

		x_deb += delta_x_deb;
		y_deb += delta_y_deb;
		th_deb += delta_th_deb;

		ROS_INFO("Before coordinate: %lf, %lf, %lf", x_deb, y_deb, th_deb);
		ROS_INFO("After coordinate: %lf, %lf, %lf", x, y, th);
		// Debug
//		ROS_INFO("delta theta: %lf, [ x: %lf, y: %lf, yaw: %lf ]", deltaTh, x, y, th);

		prevVth[2] = omega;
	}

	void updateOdom()
	{
		currentTime[0] = io->currentTime();
		double updateSince = currentTime[0] - lastTime[0];
		w_sum[0] += gyroSensor->w();
		sensingSteps[0]++;

		if(updateSince >= update_period[0]){
//			currentTime[0] = io->currentTime();
//			double deltaT = (currentTime[0] - lastTime[0]);
//			Vector2 wheel_th, wheel_omg, wheel_v;
//			Vector2 debug_omg, debug_v;
//
//			for(int i = 0; i < sizeof(wheels) / sizeof(wheels[0]); ++i){
//				double latest_wheel_angle = wheels[i]->q();
//				wheel_th[i] = (latest_wheel_angle - prevWheelTh[i]);
//				wheel_omg[i] = wheel_th[i] / update_period[0];
////				wheel_omg[i] = wheel_th[i];
//				wheel_v[i] = r * wheel_omg[i];
//				// Debug
////				debug_omg[i] = wheel_th[i] / update_period[0];
////				debug_omg[i] = wheel_th[i];
////				debug_v[i] = r * debug_omg[i];
//				prevWheelTh[i] = latest_wheel_angle;
//			}
//
//			v[0] = (wheel_v[0] + wheel_v[1]) / 2.0;
////			v[0] = (wheel_v[0] + wheel_v[1]) / (2.0 * deltaT);
//			v[1] = 0.0;
//			double omega = (wheel_v[1] - wheel_v[0]) / (2.0 * d);
////			double omega = (wheel_v[1] - wheel_v[0]) / (2.0 * d * deltaT) * 0.5;
//			// odom座標系でx, y, θ方向への速度設定
//			double deltaTh = ((omega + prevVth[2]) / 2.0) * deltaT;
//
//			// Wheel odometry + Imu
////			double deltaTh = ((gyroSensor->w().z() + prevVth[2]) / 2.0) * deltaT;
//			th += deltaTh;
//			// Debug
////			double debug_omega = (debug_v[1] - debug_v[0]) / (2.0 * d);
////			double debug_omega = (debug_v[1] - debug_v[0]) / (2.0 * d * update_period[0]);
////			double debug_dth = ((debug_omega + debug_vth[2]) / 2.0) * deltaT;
////			ROS_INFO("delta theta part1: %lf", deltaTh);
////			ROS_INFO("delta theta part2: %lf", (debug_omega + debug_vth[2]) / 2.0);
//
////			double deltaTh = ((gyroSensor->w().z() + prevVth[2]) / 2.0) * deltaT;
////			th += deltaTh;
//
//			double deltaX = (v[0] * cos(th) - v[1] * sin(th)) * deltaT;
//			double deltaY = (v[0] * sin(th) + v[1] * cos(th)) * deltaT;
////			double deltaTh = ((omega + prevVth[2]) / 2.0)  * deltaT;
//
//			x += deltaX;
//			y += deltaY;
//			ROS_INFO("x: %lf, y: %lf, theta(rad): %lf, theta(deg): %lf, omega: %lf",
//							x, y, th, th / M_PI * 180, omega);
//			th += deltaTh;
			// Debug
//			debug_th += debug_dth;
			auto w = w_sum[0] / sensingSteps[0];
			// Convert yaw value to quaternion
			geometry_msgs::Quaternion odomQuat = tf::createQuaternionMsgFromYaw(th);
//			geometry_msgs::Quaternion odomQuat = tf::createQuaternionMsgFromYaw(imuTh);
			// Create a TransformedStamped message to send to tf
			geometry_msgs::TransformStamped odomTrans;
			odomTrans.header.stamp.fromSec(currentTime[0]);
			odomTrans.header.frame_id = ODOM_FRAME_ID;
			odomTrans.child_frame_id = CHILD_FRAME_ID;

			// Substitute coordinate transformation data from odometry data and send using Transform Broadcaster
			odomTrans.transform.translation.x = x;
			odomTrans.transform.translation.y = y;
			odomTrans.transform.translation.z = 0.0;
			odomTrans.transform.rotation = odomQuat;
			tfBroadcaster_.sendTransform(odomTrans);

			// The navigation stack publishes nav_msgs / Odometry messages for velocities information
			nav_msgs::Odometry odom;
			odom.header.stamp.fromSec(currentTime[0]);
			odom.header.frame_id = ODOM_FRAME_ID;

			odom.pose.pose.position.x = x;
			odom.pose.pose.position.y = y;
			odom.pose.pose.position.z = 0.0;
			odom.pose.pose.orientation = odomQuat;

			odom.child_frame_id = CHILD_FRAME_ID;
			odom.twist.twist.linear.x = v[0];
			odom.twist.twist.linear.y = v[1];
//			odom.twist.twist.angular.z = omega;
//			odom.twist.twist.angular.z = gyroSensor->w().z();
			odom.twist.twist.angular.z = omega = w.z();

			w_sum[0].setZero();
			sensingSteps[0] = 0;

			updateCoord();

			ros::spinOnce();
			odom_pub_.publish(odom);

//			std_msgs::Float32MultiArray array;
//			array.data.resize(6);
//			array.data[0] = 0.0;
//			array.data[1] = 0.0;
//			array.data[2] = th;
//			array.data[3] = 0.0;
//			array.data[4] = 0.0;
//			array.data[5] = (th * 180) / M_PI;
//			thetaPublisher_.publish(array);

			lastTime[0] += update_period[0];
//			prevVth[2] = omega;
			// Debug
//			debug_vth[2] = debug_omega;
//			prevVth[2] = gyroSensor->w().z();

			// Debug
//			ROS_INFO("[ x y yaw ] = [ %lf %lf %lf ] ", x, y, th);

//			ROS_INFO("-------------------- Start output of odometry --------------------");
//			ROS_INFO("position: ( %lf, %lf, %lf )", x, y, 0.0);
//			ROS_INFO("orientation: ( %lf, %lf, %lf )", 0.0, 0.0, odomQuat.z);
//			ROS_INFO("linear: ( %lf, %lf, %lf )", v[0], v[1], 0.0);
//			ROS_INFO("angular: ( %lf, %lf, %lf )", 0.0, 0.0, vth[2]);
//			ROS_INFO("-------------------- End output of odometry --------------------");
		}

		std_msgs::String msg;
		msg.data = to_string(io->currentTime());
		simtime_pub_.publish(msg);
	}

	void updateImu()
	{
		currentTime[1] = io->currentTime();
		double updateSince = currentTime[1] - lastTime[1];

		dv_sum[1] += accelSensor->dv();
		w_sum[1] += gyroSensor->w();
		sensingSteps[1]++;

		if(updateSince >= update_period[1]){
			currentTime[1] = io->currentTime();
//			auto dv = accelSensor->dv();
//			auto w = gyroSensor->w();
			auto dv = dv_sum[1] / sensingSteps[1];
			auto w = w_sum[1] / sensingSteps[1];
//			omega = w.z();
			sensor_msgs::Imu imu;

//			vth[2] = w.z();
			imuTh += w.z() * update_period[1];

			geometry_msgs::Quaternion imuQuat = tf::createQuaternionMsgFromYaw(imuTh);
			imu.header.stamp.fromSec(currentTime[1]);
			imu.header.frame_id = IMU_FRAME_ID;
			imu.linear_acceleration.x = dv.x();
			imu.linear_acceleration.y = dv.y();
			imu.linear_acceleration.z = dv.z();
			imu.angular_velocity.x = w.x();
			imu.angular_velocity.y = w.y();
			imu.angular_velocity.z = w.z();
			imu.orientation = imuQuat;

			dv_sum[1].setZero();
			w_sum[1].setZero();
			sensingSteps[1] = 0;
			ros::spinOnce();
			imu_pub_.publish(imu);

			lastTime[1] += update_period[1];

//			ROS_INFO("-------------------- Start output of imu --------------------");
//			ROS_INFO("Accel Sensor  dvx: %lf, dvy: %lf, dvz: %lf", dv.x(), dv.y(), dv.z());
//			ROS_INFO("Gyro Sensor  wx: %lf, wy: %lf, wz: %lf", w.x(), w.y(), w.z());
//			ROS_INFO("-------------------- End output of imu --------------------");
		}
	}

	virtual bool control() override
	{
		current_time = io->currentTime();

		if(actuationMode == Link::JointVelocity){
			// When actuation mode is velocity.
//			static const double K = 8.8;
//			static const double K = 6.0;
//			static const double K = 30.0;
//			if(pos[1] != 0.0 && pos[0] != 0.0){
//				wheels[0]->dq_target() = -K * (pos[1] - 0.5 * pos[0]);
//				wheels[1]->dq_target() = -K * (pos[1] + 0.5 * pos[0]);

			// 按分
//			double dq_target[2], calc_diff[2];
//			dq_target[0] = clamp(cmd_vel[0], MIN_VEL, MAX_VEL) - clamp(cmd_vel_th[2], MIN_ACCEL, MAX_ACCEL) * d;
//			dq_target[1] = clamp(cmd_vel[0], MIN_VEL, MAX_VEL) + clamp(cmd_vel_th[2], MIN_ACCEL, MAX_ACCEL) * d;
//			if(dq_target[0] > MAX_VEL || dq_target[0] < MIN_VEL){
//				if(dq_target[0] > 0){
//					calc_diff[0] = dq_target[0] - MAX_VEL;
//				} else {
//					calc_diff[0] = dq_target[0] - MIN_VEL;
//				}
//				dq_target[0] = clamp(dq_target[0], MIN_VEL, MAX_VEL);
//				dq_target[1] = clamp(dq_target[1] - calc_diff[0], MIN_VEL, MAX_VEL);
//			} else if(dq_target[1] > MAX_VEL || dq_target[1] < MIN_VEL){
//				if(dq_target[1] > 0){
//					calc_diff[1] = dq_target[1] - MAX_VEL;
//				} else {
//					calc_diff[1] = dq_target[1] - MIN_VEL;
//				}
//				dq_target[0] = clamp(dq_target[0] - calc_diff[1], MIN_VEL, MAX_VEL);
//				dq_target[1] = clamp(dq_target[1], MIN_VEL, MAX_VEL);
//			}
//			wheels[0]->dq_target() = dq_target[0] / r;
//			wheels[1]->dq_target() = dq_target[1] / r;

			wheels[0]->dq_target() = 1.17 * clamp(cmd_vel[0] - clamp(cmd_vel_th[2], MIN_ACCEL, MAX_ACCEL) * d, MIN_VEL, MAX_VEL) / r;
			wheels[1]->dq_target() = 1.17 * clamp(cmd_vel[0] + clamp(cmd_vel_th[2], MIN_ACCEL, MAX_ACCEL) * d, MIN_VEL, MAX_VEL) / r;
//			wheels[0]->dq_target() = (clamp(cmd_vel[0], MIN_VEL, MAX_VEL)
//					- clamp(cmd_vel_th[2], MIN_ACCEL, MAX_ACCEL) * d) / r;
//			wheels[1]->dq_target() = (clamp(cmd_vel[0], MIN_VEL, MAX_VEL)
//					+ clamp(cmd_vel_th[2], MIN_ACCEL, MAX_ACCEL) * d) / r;
			// Debug
//			ROS_INFO("cmd_vel: [ x: %lf, th: %lf ]", cmd_vel[0], cmd_vel_th[2]);
//			updateCoord();
//			} else {
//				if(cmd_vel[0] >= 0.01 || cmd_vel_th[2] >= 0.001){
//					wheels[0]->dq_target() = 5.0 * (cmd_vel[0] / 0.26 - (cmd_vel_th[2] / 0.26) * d);
//					wheels[1]->dq_target() = 5.0 * (cmd_vel[0] / 0.26 + (cmd_vel_th[2] / 0.26) * d);
//				} else {
//					wheels[0]->dq_target() = 0.0;
//					wheels[1]->dq_target() = 0.0;
//				}
//			}
//			ROS_INFO("control function:\tv: %lf", (wheels[0]->dq_target() + wheels[1]->dq_target()) / 2.0);

		} else if(actuationMode == Link::JointTorque){
			// When actuation mode is torque.
//			static const double K = 0.016;	// rotorInertia = 0
			static const double K = 35.0;	// rotorInertia = 1.0e-06
//			static const double K = 0.15;	// cylinder
//			static const double K = 35.0;	// change the caster material
			double dq_target[2];		// Target angular velocity.
//			dq_target[0] = -K1 * pos[1] + K2 * pos[0];		// pos[1]: -1(UP), 1(DOWN)
//			dq_target[1] = -K1 * pos[1] - K2 * pos[0];		// pos[0]: -1(LEFT), 1(RIGHT)
			dq_target[0] = clamp(cmd_vel[0] - clamp(cmd_vel_th[2], MIN_ACCEL, MAX_ACCEL) * d, MIN_VEL, MAX_VEL) / r;
			dq_target[1] = clamp(cmd_vel[0] + clamp(cmd_vel_th[2], MIN_ACCEL, MAX_ACCEL) * d, MIN_VEL, MAX_VEL) / r;

			wheel_dq[0] = dq_target[0];
			wheel_dq[1] = dq_target[1];

			for(int i = 0; i < 2; ++i){
				double q = wheels[i]->q();
//				ROS_INFO("q: %lf", q);
//				if(fabs(q) < 1.0e-03){
//					q = 0.0;
//				}
				double dq = (q - qprev[i]) / dt;	// Current angular velocity.
//				ROS_INFO("i: %d, dq_target: %lf, dq: %lf", i, dq_target[i], dq);

				wheels[i]->u() = K * (dq_target[i] - dq);	// Torque value given to tire.
				qprev[i] = wheels[i]->q();	// Update previous joint angular.
			}
//			updateCoord();
		}

		updateImu();
		updateOdom();

		auto T = ioBody->rootLink()->position();
		Vector3 rpy = rpyFromRot(T.rotation());

		if(current_time - last_time >= 1.0){
			ofs << current_time << "," << T.translation().x() << ","  << T.translation().y() << "," << 0.0 << endl;
			last_time = current_time;
		}
		if((rpy[2] > 0.0 && ptheta > 0.0) || (rpy[2] < 0.0 && ptheta < 0.0)){
			count += fabs(rpy[2] - ptheta);
		} else {
			count += fabs(rpy[2] + ptheta);
		}
//		ROS_INFO("rpy[2]: %lf, ptheta: %lf", rpy[2], ptheta);
		ptheta = rpy[2];
//		ROS_INFO("actual theta: %lf", rpy[2]);
		return true;
	}

	virtual void stop() override
	{
		auto T = ioBody->rootLink()->position();
		Vector3 rpy = rpyFromRot(T.rotation());

		ROS_INFO("odom theta: %lf, actual theta: %lf", th, count);
		x = y = th = 0.0;
		for(int i = 0; i < sizeof(v)/sizeof(v[0]); ++i){
			v[i] = vth[i] = 0.0;
		}

		cmd_vel_sub_.shutdown();
		odom_pub_.shutdown();
		imu_pub_.shutdown();
		ofs.close();
	}
};

CNOID_IMPLEMENT_SIMPLE_CONTROLLER_FACTORY(TurtleBot3RosNaviController)
