/*
 * TurtleBot3RosNaviController.cpp
 *
 *  Created on: 2021/04/01
 *      Author: Tsuyoshi Anazawa
 *
 *  Outline: This program that controls the TurtleBot3 model with ROS navigation.
 */

#include <cnoid/SimpleController>
#include <cnoid/RangeCamera>
#include <cnoid/RangeSensor>
#include <cnoid/AccelerationSensor>
#include <cnoid/RateGyroSensor>
#include <fmt/format.h>
#include <ros/node_handle.h>
#include <sensor_msgs/Imu.h>
#include <mutex>
#include <iostream>
#include <ros/ros.h>
#include <nav_msgs/Odometry.h>
#include <tf/transform_broadcaster.h>
#include <geometry_msgs/Twist.h>
// simulation time
#include <std_msgs/String.h>
#include <cnoid/EigenUtil>
#include <math.h>
// Output csv file.
#include <fstream>
/* --> (Start) Add by T.Anazawa on 2022-06-01
 * Generate random numbers according to a normal distribution. */

#include "../include/NoiseGenerator.h"
/* <-- (End) Add by T.Anazawa on 2022-06-01 */
/* --> (Start) Add by T.Anazawa on 2023-09-01 */
#include <mutex>
/* <-- (End) Add by T.Anazawa on 2023-09-01 */

using namespace std;
//using namespace cnoid;
using fmt::format;

class TurtleBot3RosNaviTestController : public cnoid::SimpleController
{
	ros::NodeHandle nh_;
	// Publisher for Odometry
	ros::Publisher odom_pub_;
	tf::TransformBroadcaster tfBroadcaster_;
	// 0: odom, 1: imu, 2: cmd_vel
	double currentTime[3], lastTime[3], update_period[3];
	double update_rate[3] = { 30.0, 200.0, 10.0 };

	// Publisher for imu
	ros::Publisher imu_pub_;
	cnoid::AccelerationSensorPtr accelSensor;
	cnoid::RateGyroSensorPtr gyroSensor;
	std::string deviceName[2] = { "AccelSensor", "GyroSensor" };
	cnoid::ScopedConnection accelSensorConnection;
	cnoid::ScopedConnection gyroSensorConnection;

	// Subscriber for cmd_vel
	ros::Subscriber cmd_vel_sub_;
	// Number of wheels
	static const int WHEEL_NUM = 2;
	// Wheel names
	const std::string wheelNames[WHEEL_NUM] = { "Left_wheel", "Right_wheel" };
	// Actuation mode
	int actuationMode;
	// Wheel links
	cnoid::Link* wheels[2];
	double qprev[2];
	double dt;
	double qref[2];
	cnoid::SimpleControllerIO* io;
	// current coordinate position
	double x, y, th;
	double omega;
	// Current velocities[m/s] and angular velocities[rad/s]
	cnoid::Vector3 v, vth;
	// Last velocities[m/s] and angular velocities[rad/s]
	cnoid::Vector3 prevVth;
	double imuTh;
	cnoid::Vector3 cmd_vel, cmd_vel_th;
	double timeSpan = 0.05;
	// Odometry and imu frame id
	const std::string ODOM_FRAME_ID = "/tb3_01/odom";
	const std::string CHILD_FRAME_ID = "/tb3_01/base_footprint";
	const std::string IMU_FRAME_ID = "/tb3_01/base_footprint";
	// Limit velocities and accelerations
	const double MAX_VEL = 0.26;
	const double MIN_VEL = -0.27;
	const double MAX_ACCEL = 1.82;
	const double MIN_ACCEL = -1.82;
	// half of the tread
	double d = 0.1435;
	// Wheel radius
	double r = 0.033;
	cnoid::Vector2 prevWheelTh;
	cnoid::Vector3 dv_sum[2], w_sum[2];
	int sensingSteps[2];
	double wheel_dq[2];

	// rootLink
	cnoid::BodyPtr ioBody;
	double count = 0;
	double ptheta;
	// Simulation time.
	ros::Publisher simtime_pub_;
	// Output csv file.
	ofstream ofs;
	double current_time, last_time;
	/* <-- (Start)Add by T.Anazawa on 2023-09-01 */
	std::mutex cmdvelMutex;
	/* --> (End)Add by T.Anazawa on 2023-09-01 */

public:
	double clamp(const double x, const double min, const double max)
	{
	  return std::min(std::max(min, x), max);
	}

	virtual bool configure(cnoid::SimpleControllerConfig* config) override
	{
		return true;
	}

	virtual bool initialize(cnoid::SimpleControllerIO* io) override
	{
		this->io = io;
		ostream& os = io->os();
		// Body variable.
		cnoid::Body* body = io->body();
		// Add to actual value.
		ioBody = body;

		// Set joint's actuation mode.
		actuationMode = cnoid::Link::JointTorque;
		// Get controller option.
		std::string option = io->optionString();

		if(!option.empty()){
			// When controller option is empty.
			if(option == "velocity" || option == "position"){
				// When controller option is velocity or position.
				actuationMode = cnoid::Link::JointVelocity;
			} else if(option == "torque"){
				// When controller option is torque.
				actuationMode = cnoid::Link::JointTorque;
			} else {
				// In other cases.
				os << format("Warning: Unknown option {}.", option) << endl;
			}
		}

		// Add to actual value.
		io->enableInput(ioBody->rootLink(), LINK_POSITION);

		for(int i = 0; i < WHEEL_NUM; ++i){
			// Store wheel links in array.
			wheels[i] = body->link(wheelNames[i]);
			if(!wheels[i]){
				// When there is no wheel links.
				os << format("{0} of {1} is not found.", wheelNames[i], body->name()) << endl;
				return false;
			}

			// Set the wheel array's actuation mode.
			wheels[i]->setActuationMode(actuationMode);
			// Enable output to wheel links.
			io->enableIO(wheels[i]);
			// Store current joint angles in array
			qref[i] = qprev[i] = wheels[i]->q();
		}

		accelSensor = body->findDevice<cnoid::AccelerationSensor>(deviceName[0]);
		if(!accelSensor){
			os << format("{0} of {1} is not found.", deviceName[0], body->name()) << endl;
			//return false;
		} else {
			io->enableInput(accelSensor);
//			accelSensorConnection.disconnect();
//			accelSensorConnection = accelSensor->sigStateChanged().connect(
//					[&](){ updateImu(); });
		}
		
		gyroSensor = body->findDevice<cnoid::RateGyroSensor>(deviceName[1]);
		if(!gyroSensor){
			os << format("{0} of {1} is not found.", deviceName[1], body->name()) << endl;
			//return false;
		} else {
			io->enableInput(gyroSensor);
//			gyroSensorConnection.disconnect();
//			gyroSensorConnection = gyroSensor->sigStateChanged().connect(
//					[&](){ updateImu(); });
		}

		// Create subscriber for cmd_vel
		cmd_vel_sub_ = nh_.subscribe("cmd_vel", 1, &TurtleBot3RosNaviTestController::cmdvelCallback, this);
		// Create publisher for odometry(topic: odom, buffer_size: 10)
		odom_pub_ = nh_.advertise<nav_msgs::Odometry>("odom", 10);
		// Create publisherfor imu(topic: imu, buffer_size: 10)
		imu_pub_ = nh_.advertise<sensor_msgs::Imu>("imu", 10);
		// Create publisher for simulation time.
		simtime_pub_ = nh_.advertise<std_msgs::String>("simtime", 10);
		// Origin of odom coordinate system assuming the initial position of the robot
		nh_.param("tb3_01_x_pos", x, 0.0);
		nh_.param("tb3_01_y_pos", y, 0.0);
		nh_.param("tb3_01_yaw", th, 0.0);
		imuTh = th;
		// Add to actual value.
		ptheta = th;

		prevVth[2] = 0.0;

		for(int i = 0; i < sizeof(wheels) / sizeof(wheels[0]); ++i){
			prevWheelTh[i] = wheels[i]->q();
		}

		currentTime[3] = lastTime[3] = io->currentTime();
		for(int i = 0; i < sizeof(currentTime) / sizeof(currentTime[0]); ++i){
			update_period[i] = 1.0 / update_rate[i];
		}

		std::string file_path = "/home/fsk9001/.ros/cnoid_coord_navi.csv";
		ifstream ifs(file_path);
		std::string hdr;
		bool hdr_flg = false;
		if(ifs.fail()){
			hdr = "sim_time,pos_x,pos_y,pos_z,ori_x,ori_y,ori_z,ori_w";
			hdr_flg = true;
		}
		ofs.open(file_path, ios::app);
		if(hdr_flg){
			ofs << hdr << endl;
		}

		current_time = last_time = io->currentTime();
		dt = io->timeStep();
		sensingSteps[0] = sensingSteps[1] = 0;

		return true;
	}

	virtual bool start() override
	{
		return true;
	}

	void cmdvelCallback(const geometry_msgs::Twist& msg)
	{
		std::lock_guard<std::mutex> lock(cmdvelMutex);
		cmd_vel[0] = msg.linear.x;
		cmd_vel[1] = msg.linear.y;
		cmd_vel[2] = msg.linear.z;
		cmd_vel_th[0] = msg.angular.x;
		cmd_vel_th[1] = msg.angular.y;
		cmd_vel_th[2] = msg.angular.z;
	}

	void updateCoord()
	{
		cnoid::Vector2 wheel_th, wheel_omg, wheel_v;

		for(int i = 0; i < sizeof(wheels) / sizeof(wheels[0]); ++i){
			/* <-- (Start)Modify by T.Anazawa on 2022-06-01
			 * Change the process to get the coordinate positions. */
			/*
			double latest_wheel_angle = wheels[i]->q();
			wheel_th[i] = (latest_wheel_angle - prevWheelTh[i]);
			*/
			/* --> (End)Modify by T.Anazawa on 2022-06-01 */
//			wheel_omg[i] = wheel_th[i] / dt;
			/* <-- (Start)Modify by T.Anazawa on 2022-06-01
			 * Change the process to get the coordinate positions. */
//			wheel_omg[i] = wheel_th[i] / update_period[0];
			wheel_omg[i] = wheel_dq[i];
			/* --> (End)Modify by T.Anazawa on 2022-06-01 */

			wheel_v[i] = r * wheel_omg[i];
			/* std::string w_nm;
			if(i == 0){
				w_nm = "left";
			} else {
				w_nm = "right";
			} */
			/* <-- (Start)Modify by T.Anazawa on 2022-06-01
			 * Change the process to get the coordinate positions. */
//			prevWheelTh[i] = latest_wheel_angle;
			/* --> (End)Modify by T.Anazawa on 2022-06-01 */
		}
		v[0] = clamp((wheel_v[0] + wheel_v[1]) / 2.0, MIN_VEL, MAX_VEL);
		v[1] = 0.0;

		double deltaX = (v[0] * cos(th) - v[1] * sin(th)) * update_period[0];
		double deltaY = (v[0] * sin(th) + v[1] * cos(th)) * update_period[0];
		double deltaTh = ((omega + prevVth[2]) / 2.0) * update_period[0];

		x += deltaX;
		y += deltaY;
		th += deltaTh;

		prevVth[2] = omega;
	}

	void updateOdom()
	{
		currentTime[0] = io->currentTime();
		double updateSince = currentTime[0] - lastTime[0];
		w_sum[0] += gyroSensor->w();
		sensingSteps[0]++;

		if(updateSince >= update_period[0]){
			auto w = w_sum[0] / sensingSteps[0];
			// Convert yaw value to quaternion
			geometry_msgs::Quaternion odomQuat = tf::createQuaternionMsgFromYaw(th);
			// Create a TransformedStamped message to send to tf
			geometry_msgs::TransformStamped odomTrans;
			odomTrans.header.stamp.fromSec(currentTime[0]);
			odomTrans.header.frame_id = ODOM_FRAME_ID;
			odomTrans.child_frame_id = CHILD_FRAME_ID;

			// Substitute coordinate transformation data from odometry data and send using Transform Broadcaster
			odomTrans.transform.translation.x = x;
			odomTrans.transform.translation.y = y;
			odomTrans.transform.translation.z = 0.0;
			odomTrans.transform.rotation = odomQuat;
			tfBroadcaster_.sendTransform(odomTrans);

			// The navigation stack publishes nav_msgs / Odometry messages for velocities information
			nav_msgs::Odometry odom;
			odom.header.stamp.fromSec(currentTime[0]);
			odom.header.frame_id = ODOM_FRAME_ID;

			odom.pose.pose.position.x = x;
			odom.pose.pose.position.y = y;
			odom.pose.pose.position.z = 0.0;
			odom.pose.pose.orientation = odomQuat;

			odom.child_frame_id = CHILD_FRAME_ID;
			odom.twist.twist.linear.x = v[0];
			odom.twist.twist.linear.y = v[1];
			odom.twist.twist.angular.z = omega = w.z();

			w_sum[0].setZero();
			sensingSteps[0] = 0;

			updateCoord();

			ros::spinOnce();
			odom_pub_.publish(odom);

			lastTime[0] += update_period[0];
		}

		std_msgs::String msg;
		msg.data = std::to_string(io->currentTime());
		simtime_pub_.publish(msg);
	}

	void updateImu()
	{
		currentTime[1] = io->currentTime();
		double updateSince = currentTime[1] - lastTime[1];

		dv_sum[1] += accelSensor->dv();
		w_sum[1] += gyroSensor->w();
		sensingSteps[1]++;

		if(updateSince >= update_period[1]){
			currentTime[1] = io->currentTime();
			auto dv = dv_sum[1] / sensingSteps[1];
			auto w = w_sum[1] / sensingSteps[1];
			sensor_msgs::Imu imu;

			imuTh += w.z() * update_period[1];

			geometry_msgs::Quaternion imuQuat = tf::createQuaternionMsgFromYaw(imuTh);
			imu.header.stamp.fromSec(currentTime[1]);
			imu.header.frame_id = IMU_FRAME_ID;
			imu.linear_acceleration.x = dv.x();
			imu.linear_acceleration.y = dv.y();
			imu.linear_acceleration.z = dv.z();
			imu.angular_velocity.x = w.x();
			imu.angular_velocity.y = w.y();
			imu.angular_velocity.z = w.z();
			imu.orientation = imuQuat;

			dv_sum[1].setZero();
			w_sum[1].setZero();
			sensingSteps[1] = 0;
			ros::spinOnce();
			imu_pub_.publish(imu);

			lastTime[1] += update_period[1];
		}
	}

	/* <-- (Start)Add by T.Anazawa on 2022-06-01
	 * Add writeCsv function. */
	void writeCsv(cnoid::Isometry3& T, cnoid::Vector3 rpy)
	{
		geometry_msgs::Quaternion quat = tf::createQuaternionMsgFromRollPitchYaw(rpy[0], rpy[1], rpy[2]);

		if(current_time - last_time >= 1.0){
			ofs << current_time << "," << T.translation().x() << ","  << T.translation().y() << "," << 0.0 << ","
					<< quat.x << "," << quat.y << "," << quat.z << "," << quat.w << std::endl;
			last_time = current_time;
		}
	}
	/* --> (End)Add by T.Anazawa on 2022-06-01 */

	virtual bool control() override
	{
		current_time = io->currentTime();
		double ratio = 1.0; // 1.4;
		double dq_target[2];		// Target angular velocity.
		{
			std::lock_guard<std::mutex> lock(cmdvelMutex);
			dq_target[0] = clamp(ratio * cmd_vel[0] - clamp(ratio * cmd_vel_th[2], MIN_ACCEL, MAX_ACCEL) * d, MIN_VEL, MAX_VEL) / r;
			dq_target[1] = clamp(ratio * cmd_vel[0] + clamp(ratio * cmd_vel_th[2], MIN_ACCEL, MAX_ACCEL) * d, MIN_VEL, MAX_VEL) / r;
		}

		if(actuationMode == cnoid::Link::JointVelocity){
			// When actuation mode is velocity.
			static const double K = 1.17;
			wheels[0]->dq_target() = K * dq_target[0];
			wheels[1]->dq_target() = K * dq_target[1];

		} else if(actuationMode == cnoid::Link::JointTorque){
			// When actuation mode is torque.
			static const double K = 5.0;	// rotorInertia = 1.0e-06

			/* <-- (Start)Add by T.Anazawa on 2022-06-01
			 * Change the process to get the coordinate positions. */
			wheel_dq[0] = dq_target[0];
			wheel_dq[1] = dq_target[1];
			/* <-- (End)Add by T.Anazawa on 2022-06-01 */

			ROS_INFO("---------- wheel's dq_target ----------");
			ROS_INFO("left: %lf, right: %lf", dq_target[0], dq_target[1]);

			for(int i = 0; i < 2; ++i){
				double q = wheels[i]->q();
				double dq = (q - qprev[i]) / dt;	// Current angular velocity.

				wheels[i]->u() = K * (dq_target[i] - dq);	// Torque value given to tire.
				qprev[i] = wheels[i]->q();	// Update previous joint angular.
			}
		}

		updateImu();
		updateOdom();

//		auto T = ioBody->rootLink()->position();
//		cnoid::Vector3 rpy = cnoid::rpyFromRot(T.rotation());
//
//		/* <-- (Start)Add by T.Anazawa on 2022/06/01 Add writeCsv function. */
//		TurtleBot3RosNaviTestController::writeCsv(T, rpy);
//		/* --> (End)Add by T.Anazawa on 2022/06/01 */
//
//		if((rpy[2] > 0.0 && ptheta > 0.0) || (rpy[2] < 0.0 && ptheta < 0.0)){
//			count += std::fabs(rpy[2] - ptheta);
//		} else {
//			count += std::fabs(rpy[2] + ptheta);
//		}
//		ptheta = rpy[2];

//		Debug
//		ROS_INFO("********** Gaussian noise parameters for wheels **********");
//		ROS_INFO("detect_rate: %lf, mean: %lf, stddev: %lf", ng->detectionRate(), ng->mean(), ng->standardDeviation());

//		std::pair<double, double> p = ng->addGaussianNoise();
//		ROS_INFO("Gaussian noise for wheels: %lf", ng->addGaussianNoise());
//		ROS_INFO("********** Gaussian noise  for wheels **********");
//		ROS_INFO("rotation_error: %lf, detection_probability: %lf", p.first, p.second);

		return true;
	}

	virtual void stop() override
	{
		auto T = ioBody->rootLink()->position();
		cnoid::Vector3 rpy = cnoid::rpyFromRot(T.rotation());

		x = y = th = 0.0;
		for(int i = 0; i < sizeof(v)/sizeof(v[0]); ++i){
			v[i] = vth[i] = 0.0;
		}

		cmd_vel_sub_.shutdown();
		odom_pub_.shutdown();
		imu_pub_.shutdown();
		simtime_pub_.shutdown();
		ofs.close();
//		delete gn;
	}
};

CNOID_IMPLEMENT_SIMPLE_CONTROLLER_FACTORY(TurtleBot3RosNaviTestController)
